"""Legend generation for diagram.

This module creates a legend box on the left side of the diagram showing:
- VNet types (Hub, Spoke, External)
- Connection types (Hub-to-Spoke, Spoke-to-Spoke, Hub-to-Hub, Cross-Tenant)
- Icons (Firewall, VPN Gateway, ExpressRoute, NSG, Route Table, etc.)
"""

from xml.etree.ElementTree import Element, SubElement

from gettopology.diagram.config import get_config

# Get global config instance
_config = get_config()


def create_legend(
    root: Element,
    cell_id_counter: int,
    legend_x: float,
    legend_y: float,
) -> tuple[int, float]:
    """Create a legend box on the left side of the diagram.

    Args:
        root: Root element to add legend to
        cell_id_counter: Starting cell ID counter
        legend_x: X position for legend (left edge)
        legend_y: Y position for legend (top edge)

    Returns:
        Tuple of (updated cell_id_counter, actual_legend_height)
    """
    legend_width = _config.legend_width
    legend_height = _config.legend_height
    legend_padding = _config.legend_padding
    legend_item_spacing = _config.legend_item_spacing
    legend_item_height = _config.legend_item_height

    # Create legend container group
    legend_group = SubElement(root, "mxCell")
    legend_group.set("id", str(cell_id_counter))
    legend_group.set("value", "Legend")
    legend_group.set("style", "swimlane;whiteSpace=wrap;html=1;fillColor=#F5F5F5;strokeColor=#CCCCCC;strokeWidth=2;")
    legend_group.set("vertex", "1")
    legend_group.set("parent", "1")

    legend_geometry = SubElement(legend_group, "mxGeometry")
    legend_geometry.set("x", str(legend_x))
    legend_geometry.set("y", str(legend_y))
    legend_geometry.set("width", str(legend_width))
    legend_geometry.set("height", str(legend_height))
    legend_geometry.set("as", "geometry")

    cell_id_counter += 1

    # Calculate column widths for 3-column layout
    # Column 1: VNet Types, Column 2: Connection Types, Column 3: Icons (with 2 sub-columns)
    column_gap = _config.legend_column_gap
    available_width = legend_width - 2 * legend_padding - (2 * column_gap)  # Account for gaps between 3 columns
    column_width = available_width / 3

    # Column X positions
    col1_x = legend_x + legend_padding
    col2_x = col1_x + column_width + column_gap
    col3_x = col2_x + column_width + column_gap

    # Ensure content doesn't overflow: VNet boxes are configurable, labels start at configurable gap
    # Connection lines should be shorter to fit in column 2
    max_content_width = column_width - _config.legend_column_margin

    # Current Y position for legend items (start at top for all columns)
    current_y = legend_y + legend_padding

    # ========== COLUMN 1: VNet Types ==========
    col1_y = current_y
    section_title = SubElement(root, "mxCell")
    section_title.set("id", str(cell_id_counter))
    section_title.set("value", "VNet Types")
    section_title.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;fontStyle=1")
    section_title.set("vertex", "1")
    section_title.set("parent", "1")
    title_geom = SubElement(section_title, "mxGeometry")
    title_geom.set("x", str(col1_x))
    title_geom.set("y", str(col1_y))
    title_geom.set("width", str(column_width))
    title_geom.set("height", str(legend_item_height))
    title_geom.set("as", "geometry")
    cell_id_counter += 1
    col1_y += legend_item_height + legend_item_spacing

    # Hub (Blue Box)
    hub_box_x = col1_x
    hub_box_y = col1_y
    hub_box = SubElement(root, "mxCell")
    hub_box.set("id", str(cell_id_counter))
    hub_box.set("value", "")
    hub_box.set("style", f"rounded=0;whiteSpace=wrap;html=1;fillColor={_config.color_hub_fill};strokeColor={_config.color_hub_stroke};strokeWidth=2;")
    hub_box.set("vertex", "1")
    hub_box.set("parent", "1")
    hub_geom = SubElement(hub_box, "mxGeometry")
    hub_geom.set("x", str(hub_box_x))
    hub_geom.set("y", str(hub_box_y))
    hub_geom.set("width", str(_config.legend_vnet_box_width))
    hub_geom.set("height", str(_config.legend_vnet_box_height))
    hub_geom.set("as", "geometry")
    cell_id_counter += 1

    hub_label = SubElement(root, "mxCell")
    hub_label.set("id", str(cell_id_counter))
    hub_label.set("value", "Hub")
    hub_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    hub_label.set("vertex", "1")
    hub_label.set("parent", "1")
    hub_label_geom = SubElement(hub_label, "mxGeometry")
    hub_label_geom.set("x", str(hub_box_x + _config.legend_vnet_box_to_label_gap))
    hub_label_geom.set("y", str(hub_box_y))
    hub_label_geom.set("width", str(max_content_width - _config.legend_vnet_box_to_label_gap))
    hub_label_geom.set("height", str(_config.legend_vnet_box_height))
    hub_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col1_y += _config.legend_vnet_box_height + legend_item_spacing

    # Spoke (Orange Box)
    spoke_box_x = col1_x
    spoke_box_y = col1_y
    spoke_box = SubElement(root, "mxCell")
    spoke_box.set("id", str(cell_id_counter))
    spoke_box.set("value", "")
    spoke_box.set("style", f"rounded=0;whiteSpace=wrap;html=1;fillColor={_config.color_spoke_fill};strokeColor={_config.color_spoke_stroke};strokeWidth=2;")
    spoke_box.set("vertex", "1")
    spoke_box.set("parent", "1")
    spoke_geom = SubElement(spoke_box, "mxGeometry")
    spoke_geom.set("x", str(spoke_box_x))
    spoke_geom.set("y", str(spoke_box_y))
    spoke_geom.set("width", str(_config.legend_vnet_box_width))
    spoke_geom.set("height", str(_config.legend_vnet_box_height))
    spoke_geom.set("as", "geometry")
    cell_id_counter += 1

    spoke_label = SubElement(root, "mxCell")
    spoke_label.set("id", str(cell_id_counter))
    spoke_label.set("value", "Spoke")
    spoke_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    spoke_label.set("vertex", "1")
    spoke_label.set("parent", "1")
    spoke_label_geom = SubElement(spoke_label, "mxGeometry")
    spoke_label_geom.set("x", str(spoke_box_x + _config.legend_vnet_box_to_label_gap))
    spoke_label_geom.set("y", str(spoke_box_y))
    spoke_label_geom.set("width", str(max_content_width - _config.legend_vnet_box_to_label_gap))
    spoke_label_geom.set("height", str(_config.legend_vnet_box_height))
    spoke_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col1_y += _config.legend_vnet_box_height + legend_item_spacing

    # Stub/Orphan (Teal Box)
    stub_box_x = col1_x
    stub_box_y = col1_y
    stub_box = SubElement(root, "mxCell")
    stub_box.set("id", str(cell_id_counter))
    stub_box.set("value", "")
    stub_box.set("style", f"rounded=0;whiteSpace=wrap;html=1;fillColor={_config.color_stub_fill};strokeColor={_config.color_stub_stroke};strokeWidth=2;")
    stub_box.set("vertex", "1")
    stub_box.set("parent", "1")
    stub_geom = SubElement(stub_box, "mxGeometry")
    stub_geom.set("x", str(stub_box_x))
    stub_geom.set("y", str(stub_box_y))
    stub_geom.set("width", str(_config.legend_vnet_box_width))
    stub_geom.set("height", str(_config.legend_vnet_box_height))
    stub_geom.set("as", "geometry")
    cell_id_counter += 1

    stub_label = SubElement(root, "mxCell")
    stub_label.set("id", str(cell_id_counter))
    stub_label.set("value", "Stub/Orphan")
    stub_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    stub_label.set("vertex", "1")
    stub_label.set("parent", "1")
    stub_label_geom = SubElement(stub_label, "mxGeometry")
    stub_label_geom.set("x", str(stub_box_x + _config.legend_vnet_box_to_label_gap))
    stub_label_geom.set("y", str(stub_box_y))
    stub_label_geom.set("width", str(max_content_width - _config.legend_vnet_box_to_label_gap))
    stub_label_geom.set("height", str(_config.legend_vnet_box_height))
    stub_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col1_y += _config.legend_vnet_box_height + legend_item_spacing

    # External/Cross Tenant (Gray dashed box)
    external_box_x = col1_x
    external_box_y = col1_y
    external_box = SubElement(root, "mxCell")
    external_box.set("id", str(cell_id_counter))
    external_box.set("value", "")
    external_box.set("style", f"rounded=0;whiteSpace=wrap;html=1;fillColor={_config.color_external_fill};strokeColor={_config.color_external_stroke};strokeWidth=2;dashed=1;dashPattern=8 8;")
    external_box.set("vertex", "1")
    external_box.set("parent", "1")
    external_geom = SubElement(external_box, "mxGeometry")
    external_geom.set("x", str(external_box_x))
    external_geom.set("y", str(external_box_y))
    external_geom.set("width", str(_config.legend_vnet_box_width))
    external_geom.set("height", str(_config.legend_vnet_box_height))
    external_geom.set("as", "geometry")
    cell_id_counter += 1

    external_label = SubElement(root, "mxCell")
    external_label.set("id", str(cell_id_counter))
    external_label.set("value", "Cross Tenant Peer")
    external_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    external_label.set("vertex", "1")
    external_label.set("parent", "1")
    external_label_geom = SubElement(external_label, "mxGeometry")
    external_label_geom.set("x", str(external_box_x + _config.legend_vnet_box_to_label_gap))
    external_label_geom.set("y", str(external_box_y))
    external_label_geom.set("width", str(max_content_width - _config.legend_vnet_box_to_label_gap))
    external_label_geom.set("height", str(_config.legend_vnet_box_height))
    external_label_geom.set("as", "geometry")
    cell_id_counter += 1

    # ========== COLUMN 2: Connection Types ==========
    col2_y = current_y
    section_title = SubElement(root, "mxCell")
    section_title.set("id", str(cell_id_counter))
    section_title.set("value", "Connection Types")
    section_title.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;fontStyle=1")
    section_title.set("vertex", "1")
    section_title.set("parent", "1")
    title_geom = SubElement(section_title, "mxGeometry")
    title_geom.set("x", str(col2_x))
    title_geom.set("y", str(col2_y))
    title_geom.set("width", str(column_width))
    title_geom.set("height", str(legend_item_height))
    title_geom.set("as", "geometry")
    cell_id_counter += 1
    col2_y += legend_item_height + legend_item_spacing

    # Create sample connection lines for legend
    # Need to create small invisible boxes as endpoints for edges
    # Make lines shorter to fit within column 2 width
    line_length = min(_config.legend_connection_line_length, max_content_width - _config.legend_connection_line_to_label_gap - 5)
    line_start_x = col2_x
    line_end_x = col2_x + line_length
    line_y = col2_y + (_config.legend_vnet_box_height / 2)  # Middle of item height

    # Helper function to create a connection line with invisible endpoints
    def create_legend_line(start_x, end_x, y_pos, style, line_id):
        source_id = line_id
        # Create invisible source box (completely transparent)
        source_box = SubElement(root, "mxCell")
        source_box.set("id", str(source_id))
        source_box.set("value", "")
        source_box.set("style", "points=[];outlineConnect=0;fillColor=none;strokeColor=none;opacity=0;")
        source_box.set("vertex", "1")
        source_box.set("parent", "1")
        source_geom = SubElement(source_box, "mxGeometry")
        source_geom.set("x", str(start_x))
        source_geom.set("y", str(y_pos - 1))
        source_geom.set("width", "1")
        source_geom.set("height", "1")
        source_geom.set("as", "geometry")
        line_id += 1

        target_id = line_id
        # Create invisible target box (completely transparent)
        target_box = SubElement(root, "mxCell")
        target_box.set("id", str(target_id))
        target_box.set("value", "")
        target_box.set("style", "points=[];outlineConnect=0;fillColor=none;strokeColor=none;opacity=0;")
        target_box.set("vertex", "1")
        target_box.set("parent", "1")
        target_geom = SubElement(target_box, "mxGeometry")
        target_geom.set("x", str(end_x))
        target_geom.set("y", str(y_pos - 1))
        target_geom.set("width", "1")
        target_geom.set("height", "1")
        target_geom.set("as", "geometry")
        line_id += 1

        # Create edge connecting the boxes
        edge = SubElement(root, "mxCell")
        edge.set("id", str(line_id))
        edge.set("value", "")
        edge.set("style", style)
        edge.set("edge", "1")
        edge.set("parent", "1")
        edge.set("source", str(source_id))  # Source box ID
        edge.set("target", str(target_id))  # Target box ID
        edge_geom = SubElement(edge, "mxGeometry")
        edge_geom.set("relative", "1")
        edge_geom.set("as", "geometry")
        line_id += 1

        return line_id

    # Maroon dashed line - Spoke to Spoke
    cell_id_counter = create_legend_line(
        line_start_x, line_end_x, line_y,
        f"edgeStyle=orthogonalEdgeStyle;rounded=0;html=1;strokeColor={_config.edge_color_hubless_spoke};strokeWidth=2;dashed=1;dashPattern=8 8;startArrow=block;endArrow=block;",
        cell_id_counter
    )

    spoke_to_spoke_label = SubElement(root, "mxCell")
    spoke_to_spoke_label.set("id", str(cell_id_counter))
    spoke_to_spoke_label.set("value", "Spoke to Spoke")
    spoke_to_spoke_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    spoke_to_spoke_label.set("vertex", "1")
    spoke_to_spoke_label.set("parent", "1")
    spoke_to_spoke_label_geom = SubElement(spoke_to_spoke_label, "mxGeometry")
    spoke_to_spoke_label_geom.set("x", str(line_end_x + _config.legend_connection_line_to_label_gap))
    spoke_to_spoke_label_geom.set("y", str(col2_y))
    spoke_to_spoke_label_geom.set("width", str(max_content_width - (line_end_x - col2_x) - _config.legend_connection_line_to_label_gap))
    spoke_to_spoke_label_geom.set("height", str(_config.legend_vnet_box_height))
    spoke_to_spoke_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col2_y += _config.legend_vnet_box_height + legend_item_spacing

    # Blue line - Hub to Spoke
    line_y = col2_y + (_config.legend_vnet_box_height / 2)
    cell_id_counter = create_legend_line(
        line_start_x, line_end_x, line_y,
        f"edgeStyle=orthogonalEdgeStyle;rounded=0;html=1;strokeColor={_config.color_hub_to_spoke};strokeWidth=2;startArrow=block;endArrow=block;",
        cell_id_counter
    )

    hub_to_spoke_label = SubElement(root, "mxCell")
    hub_to_spoke_label.set("id", str(cell_id_counter))
    hub_to_spoke_label.set("value", "Hub to Spoke")
    hub_to_spoke_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    hub_to_spoke_label.set("vertex", "1")
    hub_to_spoke_label.set("parent", "1")
    hub_to_spoke_label_geom = SubElement(hub_to_spoke_label, "mxGeometry")
    hub_to_spoke_label_geom.set("x", str(line_end_x + _config.legend_connection_line_to_label_gap))
    hub_to_spoke_label_geom.set("y", str(col2_y))
    hub_to_spoke_label_geom.set("width", str(max_content_width - (line_end_x - col2_x) - _config.legend_connection_line_to_label_gap))
    hub_to_spoke_label_geom.set("height", str(_config.legend_vnet_box_height))
    hub_to_spoke_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col2_y += _config.legend_vnet_box_height + legend_item_spacing

    # Green line - Hub to Hub
    line_y = col2_y + (_config.legend_vnet_box_height / 2)
    cell_id_counter = create_legend_line(
        line_start_x, line_end_x, line_y,
        f"edgeStyle=orthogonalEdgeStyle;rounded=0;html=1;strokeColor={_config.color_hub_to_hub};strokeWidth=2;startArrow=block;endArrow=block;",
        cell_id_counter
    )

    hub_to_hub_label = SubElement(root, "mxCell")
    hub_to_hub_label.set("id", str(cell_id_counter))
    hub_to_hub_label.set("value", "Hub to Hub")
    hub_to_hub_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    hub_to_hub_label.set("vertex", "1")
    hub_to_hub_label.set("parent", "1")
    hub_to_hub_label_geom = SubElement(hub_to_hub_label, "mxGeometry")
    hub_to_hub_label_geom.set("x", str(line_end_x + _config.legend_connection_line_to_label_gap))
    hub_to_hub_label_geom.set("y", str(col2_y))
    hub_to_hub_label_geom.set("width", str(max_content_width - (line_end_x - col2_x) - _config.legend_connection_line_to_label_gap))
    hub_to_hub_label_geom.set("height", str(_config.legend_vnet_box_height))
    hub_to_hub_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col2_y += _config.legend_vnet_box_height + legend_item_spacing

    # Gray line - Cross Tenant
    line_y = col2_y + (_config.legend_vnet_box_height / 2)
    cell_id_counter = create_legend_line(
        line_start_x, line_end_x, line_y,
        f"edgeStyle=orthogonalEdgeStyle;rounded=0;html=1;strokeColor={_config.edge_color_cross_tenant};strokeWidth=2;startArrow=block;endArrow=block;",
        cell_id_counter
    )

    cross_tenant_label = SubElement(root, "mxCell")
    cross_tenant_label.set("id", str(cell_id_counter))
    cross_tenant_label.set("value", "Cross Tenant")
    cross_tenant_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    cross_tenant_label.set("vertex", "1")
    cross_tenant_label.set("parent", "1")
    cross_tenant_label_geom = SubElement(cross_tenant_label, "mxGeometry")
    cross_tenant_label_geom.set("x", str(line_end_x + _config.legend_connection_line_to_label_gap))
    cross_tenant_label_geom.set("y", str(col2_y))
    cross_tenant_label_geom.set("width", str(max_content_width - (line_end_x - col2_x) - _config.legend_connection_line_to_label_gap))
    cross_tenant_label_geom.set("height", str(_config.legend_vnet_box_height))
    cross_tenant_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col2_y += _config.legend_vnet_box_height + legend_item_spacing

    # Green solid line - Hybrid Connected
    line_y = col2_y + (_config.legend_vnet_box_height / 2)
    cell_id_counter = create_legend_line(
        line_start_x, line_end_x, line_y,
        "edgeStyle=orthogonalEdgeStyle;rounded=0;html=1;strokeColor=#28A745;strokeWidth=3;startArrow=block;endArrow=block;",
        cell_id_counter
    )

    hybrid_connected_label = SubElement(root, "mxCell")
    hybrid_connected_label.set("id", str(cell_id_counter))
    hybrid_connected_label.set("value", "Hybrid Connected")
    hybrid_connected_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    hybrid_connected_label.set("vertex", "1")
    hybrid_connected_label.set("parent", "1")
    hybrid_connected_label_geom = SubElement(hybrid_connected_label, "mxGeometry")
    hybrid_connected_label_geom.set("x", str(line_end_x + _config.legend_connection_line_to_label_gap))
    hybrid_connected_label_geom.set("y", str(col2_y))
    hybrid_connected_label_geom.set("width", str(max_content_width - (line_end_x - col2_x) - _config.legend_connection_line_to_label_gap))
    hybrid_connected_label_geom.set("height", str(_config.legend_vnet_box_height))
    hybrid_connected_label_geom.set("as", "geometry")
    cell_id_counter += 1
    col2_y += _config.legend_vnet_box_height + legend_item_spacing

    # Red dotted line - Hybrid Not Connected
    line_y = col2_y + (_config.legend_vnet_box_height / 2)
    cell_id_counter = create_legend_line(
        line_start_x, line_end_x, line_y,
        "edgeStyle=orthogonalEdgeStyle;rounded=0;html=1;strokeColor=#FF6B6B;strokeWidth=1;dashed=1;dashPattern=8 8;startArrow=block;endArrow=block;",
        cell_id_counter
    )

    hybrid_not_connected_label = SubElement(root, "mxCell")
    hybrid_not_connected_label.set("id", str(cell_id_counter))
    hybrid_not_connected_label.set("value", "Hybrid Not Connected")
    hybrid_not_connected_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
    hybrid_not_connected_label.set("vertex", "1")
    hybrid_not_connected_label.set("parent", "1")
    hybrid_not_connected_label_geom = SubElement(hybrid_not_connected_label, "mxGeometry")
    hybrid_not_connected_label_geom.set("x", str(line_end_x + _config.legend_connection_line_to_label_gap))
    hybrid_not_connected_label_geom.set("y", str(col2_y))
    hybrid_not_connected_label_geom.set("width", str(max_content_width - (line_end_x - col2_x) - _config.legend_connection_line_to_label_gap))
    hybrid_not_connected_label_geom.set("height", str(_config.legend_vnet_box_height))
    hybrid_not_connected_label_geom.set("as", "geometry")
    cell_id_counter += 1

    # ========== COLUMN 3: Icons ==========
    col3_y = current_y
    section_title = SubElement(root, "mxCell")
    section_title.set("id", str(cell_id_counter))
    section_title.set("value", "Icons")
    section_title.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;fontStyle=1")
    section_title.set("vertex", "1")
    section_title.set("parent", "1")
    title_geom = SubElement(section_title, "mxGeometry")
    title_geom.set("x", str(col3_x))
    title_geom.set("y", str(col3_y))
    title_geom.set("width", str(column_width))
    title_geom.set("height", str(legend_item_height))
    title_geom.set("as", "geometry")
    cell_id_counter += 1
    col3_y += legend_item_height + legend_item_spacing

    # Add icon legend items in 2 columns to save vertical space
    icons = [
        ("Firewall", _config.icon_firewall),
        ("Bastion", _config.icon_bastion),
        ("VPN Gateway", _config.icon_vpn_gateway),
        ("ExpressRoute", _config.icon_expressroute),
        ("NSG", _config.icon_nsg),
        ("Route Table", _config.icon_route_table),
        ("NAT", _config.icon_nat),
        ("Private Endpoint", _config.icon_private_endpoint),
        ("DDoS Protection", _config.icon_ddos),
    ]

    icon_size = _config.legend_icon_size
    icons_per_column = _config.legend_icons_per_sub_column
    icon_sub_column_gap = _config.legend_icon_sub_column_gap
    icon_sub_column_width = (max_content_width - icon_sub_column_gap) / 2  # Divide column 3 width between 2 sub-columns

    # Track Y position for each sub-column within column 3
    icon_sub_col1_x = col3_x
    icon_sub_col2_x = col3_x + icon_sub_column_width + icon_sub_column_gap
    icon_column_y_positions = [col3_y, col3_y]  # Start both sub-columns at same Y

    for index, (icon_name, icon_path) in enumerate(icons):
        # Determine which column (0 or 1)
        column = 0 if index < icons_per_column else 1
        index if column == 0 else index - icons_per_column

        # Calculate X position based on sub-column within column 3
        icon_x = icon_sub_col1_x if column == 0 else icon_sub_col2_x
        icon_y = icon_column_y_positions[column]

        # Icon image
        icon_cell = SubElement(root, "mxCell")
        icon_cell.set("id", str(cell_id_counter))
        icon_cell.set("value", "")
        icon_cell.set("style", f"shape=image;verticalLabelPosition=bottom;verticalAlign=top;imageAspect=0;aspect=fixed;image={icon_path};")
        icon_cell.set("vertex", "1")
        icon_cell.set("parent", "1")
        icon_geom = SubElement(icon_cell, "mxGeometry")
        icon_geom.set("x", str(icon_x))
        icon_geom.set("y", str(icon_y))
        icon_geom.set("width", str(icon_size))
        icon_geom.set("height", str(icon_size))
        icon_geom.set("as", "geometry")
        cell_id_counter += 1

        # Icon label
        icon_label = SubElement(root, "mxCell")
        icon_label.set("id", str(cell_id_counter))
        icon_label.set("value", icon_name)
        icon_label.set("style", "text;html=1;strokeColor=none;fillColor=none;align=left;verticalAlign=middle;whiteSpace=wrap;rounded=0;")
        icon_label.set("vertex", "1")
        icon_label.set("parent", "1")
        icon_label_geom = SubElement(icon_label, "mxGeometry")
        icon_label_geom.set("x", str(icon_x + icon_size + _config.legend_icon_to_label_gap))
        icon_label_geom.set("y", str(icon_y))
        icon_label_geom.set("width", str(icon_sub_column_width - icon_size - _config.legend_icon_to_label_gap))
        icon_label_geom.set("height", str(icon_size))
        icon_label_geom.set("as", "geometry")
        cell_id_counter += 1

        # Update Y position for this sub-column
        icon_column_y_positions[column] += icon_size + legend_item_spacing

    # Update current_y to the maximum Y position across all columns
    current_y = max(col1_y, col2_y, max(icon_column_y_positions))

    # Calculate actual height needed and update legend box if needed
    actual_height = current_y - legend_y + legend_padding
    if actual_height > legend_height:
        # Update legend box height to fit content
        legend_geometry.set("height", str(actual_height))
    else:
        # Use configured height if content fits
        actual_height = legend_height

    return cell_id_counter, actual_height




"""Azure service operations for querying subscriptions and network resources."""

import json
import logging
import subprocess
import time
from collections.abc import Callable
from datetime import datetime
from functools import wraps
from typing import Any, TypeVar

from azure.core.exceptions import AzureError, ServiceResponseError
from azure.mgmt.resource import SubscriptionClient
from azure.mgmt.resourcegraph import ResourceGraphClient
from azure.mgmt.resourcegraph.models import QueryRequest

from gettopology.models import (
    ConnectionModel,
    LocalNetworkGatewayModel,
    TopologyModel,
    VirtualNetworkModel,
)

T = TypeVar('T')


def retry_with_backoff(
    max_retries: int = 2,
    initial_delay: float = 1.0,
    backoff_factor: float = 2.0,
    retryable_exceptions: tuple[type[Exception], ...] = (ServiceResponseError, ConnectionError, TimeoutError),
):
    """Decorator to retry function calls with exponential backoff.

    Args:
        max_retries: Maximum number of retry attempts (default: 2, so 3 total attempts)
        initial_delay: Initial delay in seconds before first retry (default: 1.0)
        backoff_factor: Multiplier for delay between retries (default: 2.0)
        retryable_exceptions: Tuple of exception types to retry on

    Returns:
        Decorated function that retries on specified exceptions
    """
    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            delay = initial_delay
            last_exception = None

            for attempt in range(max_retries + 1):  # +1 for initial attempt
                try:
                    return func(*args, **kwargs)
                except retryable_exceptions as e:
                    last_exception = e
                    if attempt < max_retries:
                        logging.warning(
                            f"Attempt {attempt + 1}/{max_retries + 1} failed for {func.__name__}: {e}. "
                            f"Retrying in {delay:.1f} seconds..."
                        )
                        time.sleep(delay)
                        delay *= backoff_factor
                    else:
                        logging.error(
                            f"All {max_retries + 1} attempts failed for {func.__name__}: {e}"
                        )
                except Exception:
                    # Don't retry on non-retryable exceptions
                    raise

            # If we get here, all retries failed
            raise last_exception  # type: ignore

        return wrapper
    return decorator


def _get_subscriptions_via_cli() -> list[str]:
    """Get all subscription IDs via Azure CLI (supports multi-tenant).

    This is a fallback method that uses 'az account list --all' to get subscriptions
    across all tenants that the user has authenticated to.

    Returns:
        List of subscription IDs, or empty list if Azure CLI is not available
    """
    try:
        result = subprocess.run(
            ["az", "account", "list", "--all", "--output", "json"],
            capture_output=True,
            text=True,
            timeout=10,
            check=False
        )
        if result.returncode == 0:
            accounts = json.loads(result.stdout)
            subscription_ids = [acc.get("id") for acc in accounts if acc.get("id")]
            if subscription_ids:
                logging.debug(f"Found {len(subscription_ids)} subscription(s) via Azure CLI across all tenants")
            return subscription_ids
        return []
    except (subprocess.TimeoutExpired, FileNotFoundError, json.JSONDecodeError, Exception) as e:
        logging.debug(f"Could not get subscriptions via Azure CLI: {e}")
        return []


def get_all_subscription_ids(credential: Any) -> list[str]:
    """Get all subscription IDs accessible to the authenticated user/service principal.

    Tries multiple methods in order:
    1. SubscriptionClient (default tenant)
    2. Azure CLI fallback (if available, for multi-tenant)
    3. Individual subscription queries (if specific subscriptions are known)

    Args:
        credential: Azure credential object from azure_auth.get_azure_credential()

    Returns:
        List of subscription IDs (UUIDs)

    Raises:
        AzureError: If unable to query subscriptions
    """
    try:
        # Use SubscriptionClient to list all accessible subscriptions
        subscription_client = SubscriptionClient(credential=credential)

        @retry_with_backoff(max_retries=2)
        def _list_subscriptions():
            return list(subscription_client.subscriptions.list())

        subscriptions = _list_subscriptions()
        subscription_ids = [sub.subscription_id for sub in subscriptions]

        if not subscription_ids:
            logging.warning("No subscriptions found via SubscriptionClient, trying Azure CLI fallback...")
            # Fallback to Azure CLI for multi-tenant support (if available)
            cli_subscriptions = _get_subscriptions_via_cli()
            if cli_subscriptions:
                subscription_ids = cli_subscriptions
                logging.info(f"Found {len(subscription_ids)}:{subscription_ids} accessible subscription(s) via Azure CLI (multi-tenant)")
                return subscription_ids
            logging.warning("No subscriptions found for the authenticated account")
            return []

        # logging.info(f"Found {len(subscription_ids)}:{subscription_ids} accessible subscription(s) in the Tenant")
        return subscription_ids

    except AzureError as e:
        logging.warning(f"SubscriptionClient failed: {e}, trying Azure CLI fallback...")
        # Fallback to Azure CLI for multi-tenant support (if available)
        cli_subscriptions = _get_subscriptions_via_cli()
        if cli_subscriptions:
            logging.info(f"Found {len(cli_subscriptions)} subscription(s) via Azure CLI (multi-tenant)")
            return cli_subscriptions
        logging.error(f"Failed to retrieve subscriptions: {e}")
        raise
    except Exception as e:
        logging.warning(f"Unexpected error with SubscriptionClient: {e}, trying Azure CLI fallback...")
        # Fallback to Azure CLI for multi-tenant support (if available)
        cli_subscriptions = _get_subscriptions_via_cli()
        if cli_subscriptions:
            logging.info(f"Found {len(cli_subscriptions)} subscription(s) via Azure CLI (multi-tenant)")
            return cli_subscriptions
        logging.error(f"Unexpected error retrieving subscriptions: {e}")
        raise AzureError(f"Failed to retrieve subscriptions: {e}") from e


def get_subscription_name(subscription_id: str, credential: Any) -> str | None:
    """Get subscription name from subscription ID.

    Args:
        subscription_id: Subscription ID (UUID)
        credential: Azure credential object

    Returns:
        Subscription name if found, None otherwise

    Raises:
        AzureError: If unable to query subscription
    """
    try:
        subscription_client = SubscriptionClient(credential=credential)

        @retry_with_backoff(max_retries=2)
        def _get_subscription():
            return subscription_client.subscriptions.get(subscription_id)

        subscription = _get_subscription()
        return subscription.display_name if subscription else None
    except AzureError as e:
        logging.warning(f"Failed to get subscription name for {subscription_id}: {e}")
        return None
    except Exception as e:
        logging.warning(f"Unexpected error getting subscription name for {subscription_id}: {e}")
        return None


def get_subscription_names(subscription_ids: list[str], credential: Any) -> dict[str, str]:
    """Get subscription names for multiple subscription IDs.

    Args:
        subscription_ids: List of subscription IDs
        credential: Azure credential object

    Returns:
        Dictionary mapping subscription_id -> subscription_name
    """
    subscription_names: dict[str, str] = {}

    # Get all subscriptions at once for efficiency
    try:
        subscription_client = SubscriptionClient(credential=credential)

        @retry_with_backoff(max_retries=2)
        def _list_subscriptions():
            return list(subscription_client.subscriptions.list())

        subscriptions = _list_subscriptions()

        # Build a map of subscription_id -> display_name
        for sub in subscriptions:
            if sub.subscription_id in subscription_ids:
                subscription_names[sub.subscription_id] = sub.display_name

        # Fill in None for any subscriptions not found
        for sub_id in subscription_ids:
            if sub_id not in subscription_names:
                subscription_names[sub_id] = None  # type: ignore

        return subscription_names

    except AzureError as e:
        logging.warning(f"Failed to get subscription names: {e}")
        # Return empty dict, subscription_name will be None
        return {}
    except Exception as e:
        logging.warning(f"Unexpected error getting subscription names: {e}")
        return {}


def validate_subscription_access(
    subscription_ids: list[str],
    credential: Any,
    ) -> tuple[list[str], list[str]]:
    """Validate that subscription IDs exist and are accessible to the authenticated user/service principal.

    This function validates subscriptions directly by attempting to access each one,
    rather than fetching all subscriptions first. This is more efficient when
    explicit subscriptions are provided via -s or -f flags.

    Args:
        subscription_ids: List of subscription IDs to validate
        credential: Azure credential object from azure_auth.get_azure_credential()

    Returns:
        tuple: (accessible_subscription_ids, inaccessible_subscription_ids)

    Raises:
        AzureError: If unable to query subscriptions
    """
    try:
        accessible = []
        inaccessible = []

        subscription_client = SubscriptionClient(credential=credential)

        # Validate each subscription directly by attempting to access it
        # This avoids fetching all subscriptions when explicit ones are provided
        for sub_id in subscription_ids:
            try:
                @retry_with_backoff(max_retries=2)
                def _get_subscription():
                    return subscription_client.subscriptions.get(sub_id)

                # Try to get subscription details - this validates access
                _get_subscription()
                accessible.append(sub_id)
            except AzureError as e:
                # Subscription not accessible or doesn't exist
                error_msg = str(e).lower()
                error_code = getattr(e, 'status_code', None) or getattr(e, 'error_code', None)

                # Log detailed error for debugging
                logging.debug(f"Error accessing subscription {sub_id}: {e} (code: {error_code})")

                # Check if it's a tenant-related error (subscription in different tenant)
                if any(keyword in error_msg for keyword in ["tenant", "invalidauthenticationtokentenant"]):
                    # Try Azure CLI as last resort (if available) - might show subscription exists but in different tenant
                    cli_subscriptions = _get_subscriptions_via_cli()
                    if sub_id in cli_subscriptions:
                        # Subscription exists but is in a different tenant
                        logging.warning(
                            f"Subscription {sub_id} exists but is in a different tenant. "
                            f"Azure CLI and Service Principal authentication can only access one tenant at a time."
                        )
                    else:
                        logging.warning(
                            f"Subscription {sub_id} may be in a different tenant or does not exist. "
                            f"Error: {str(e)[:200]}..."
                        )
                        logging.debug(f"Full error details: {e}")
                else:
                    # Other error (permissions, not found, etc.)
                    logging.debug(f"Subscription {sub_id} not accessible: {e}")

                inaccessible.append(sub_id)
            except Exception as e:
                # Unexpected error - treat as inaccessible
                logging.warning(f"Error checking subscription {sub_id}: {e}")
                inaccessible.append(sub_id)

        # Don't log here - let the caller handle logging based on context
        # This avoids duplicate log messages

        return accessible, inaccessible

    except AzureError as e:
        logging.error(f"Failed to validate subscription access: {e}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error validating subscription access: {e}")
        raise AzureError(f"Failed to validate subscription access: {e}") from e


def check_subscription_roles(
    subscription_ids: list[str],
    credential: Any,
    ) -> tuple[list[str], list[str]]:
    """Check if the authenticated user/service principal has at least Reader role on subscriptions.

    This function verifies permissions by attempting to read subscription details.
    If successful, the principal has at least Reader role (or Contributor/Owner).

    Args:
        subscription_ids: List of subscription IDs to check
        credential: Azure credential object from azure_auth.get_azure_credential()

    Returns:
        tuple: (subscriptions_with_reader_role: list[str], subscriptions_without_reader_role: list[str])

    Raises:
        AzureError: If unable to check roles
    """
    subscriptions_with_role: list[str] = []
    subscriptions_without_role: list[str] = []

    try:
        # Use SubscriptionClient to check if we can read subscription details
        # This requires Reader role or above
        subscription_client = SubscriptionClient(credential=credential)

        for subscription_id in subscription_ids:
            try:
                @retry_with_backoff(max_retries=2)
                def _get_subscription():
                    return subscription_client.subscriptions.get(subscription_id)

                # Try to get subscription details - this requires Reader role or above
                # If this succeeds, we have at least Reader role
                _get_subscription()

                # Success - we have at least Reader role
                subscriptions_with_role.append(subscription_id)
                # logging.info(f"Subscription {subscription_id}: Has at least Reader role")

            except AzureError as e:
                # If we can't read subscription, we likely don't have Reader role
                # or the subscription doesn't exist
                subscriptions_without_role.append(subscription_id)
                logging.warning(
                    f"Subscription {subscription_id}: Does not have Reader role or above. Error: {e}"
                )
            except Exception as e:
                logging.warning(f"Unexpected error checking roles for subscription {subscription_id}: {e}")
                subscriptions_without_role.append(subscription_id)

        return subscriptions_with_role, subscriptions_without_role

    except Exception as e:
        logging.error(f"Failed to check subscription roles: {e}")
        raise AzureError(f"Failed to check subscription roles: {e}") from e


def get_vnets(
    vnet_names: list[str] | None = None,
    subscription_ids: list[str] | None = None,
    credential: Any = None,
    ) -> list[str]:
    """Get virtual network names from Azure.

    Optimization strategy:
    - Always uses Azure Resource Graph (ARG) for VNet discovery/validation
      (single query across all subscriptions is much faster than iterating with NetworkManagementClient)
    - If vnet_names provided: ARG query filters by names, validates they exist
    - If vnet_names not provided: ARG query returns all VNets

    Args:
        vnet_names: List of VNet names to validate (optional). If provided, validates they exist.
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object (required)

    Returns:
        List of VNet names:
        - If vnet_names provided: Returns only the VNets that exist (drops non-existent ones)
        - If vnet_names is None: Returns all VNets found in the specified subscriptions

    Raises:
        ValueError: If credential is required but not provided
        AzureError: If Azure API calls fail
    """
    if credential is None:
        raise ValueError("credential is required")

    # If no subscription_ids provided, get all subscriptions
    if not subscription_ids:
        if vnet_names is not None:
            logging.info("No subscriptions provided, getting all accessible subscriptions for VNet validation")
        else:
            logging.info("No subscriptions provided, getting all accessible subscriptions to list all VNets")
        subscription_ids = get_all_subscription_ids(credential)
        if not subscription_ids:
            logging.warning("No subscriptions found")
            return []

    # Optimization: Always use ARG for VNet discovery/validation
    # ARG is faster than NetworkManagementClient because it's a single query across all subscriptions
    # vs iterating each subscription with NetworkManagementClient
    if vnet_names is not None:
        # VNet names provided: Use ARG to validate (single query, filter by names)
        return _get_vnets_via_arg(vnet_names, subscription_ids, credential)
    else:
        # No VNet names provided: Use ARG to get all VNets (single query)
        return _get_all_vnets_via_arg(subscription_ids, credential)


def _get_vnets_via_arg(
    vnet_names: list[str],
    subscription_ids: list[str],
    credential: Any,
    ) -> list[str]:
    """Validate specific VNet names using Azure Resource Graph.

    This is optimized for when VNet names are provided. ARG allows a single
    query across all subscriptions to validate VNets, which is much faster
    than iterating with NetworkManagementClient (which requires subscription ID per call).

    Args:
        vnet_names: List of VNet names to validate
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        List of validated VNet names that exist in the subscriptions
    """
    vnet_names_set = set(vnet_names)
    found_vnets: list[str] = []

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Build ARG query to find VNets by name across all subscriptions
        # Query returns: name, subscriptionId, resourceGroup, id
        # Use 'or' conditions for name matching (KQL syntax for ARG)
        name_conditions = " or ".join([f"name == '{name}'" for name in vnet_names])
        query = f"""
        Resources
        | where type == 'microsoft.network/virtualnetworks'
        | where {name_conditions}
        | project name, subscriptionId, resourceGroup, id
        """

        query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=query
        )

        @retry_with_backoff(max_retries=2)
        def _query_resources():
            return resource_graph_client.resources(query_request)

        response = _query_resources()

        if response.data:
            for vnet_data in response.data:
                vnet_name = vnet_data.get('name')
                if vnet_name and vnet_name in vnet_names_set and vnet_name not in found_vnets:
                    found_vnets.append(vnet_name)

        # Log missing VNets at INFO level
        missing_vnets = [vnet for vnet in vnet_names if vnet not in found_vnets]
        if missing_vnets:
            logging.info(
                f"VNet name(s) not found in specified subscriptions: {', '.join(missing_vnets)}"
            )

        if found_vnets:
            logging.info(
                f"Found {len(found_vnets)} VNet(s) in specified subscriptions (via ARG): {', '.join(found_vnets)}"
            )
        else:
            logging.warning(
                f"None of the provided VNet names found in subscriptions: {', '.join(vnet_names)}"
            )

        return found_vnets

    except AzureError as e:
        logging.error(f"Failed to query virtual networks via ARG: {e}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error querying virtual networks via ARG: {e}")
        raise AzureError(f"Failed to query virtual networks via ARG: {e}") from e


def _get_all_vnets_via_arg(
    subscription_ids: list[str],
    credential: Any,
    ) -> list[str]:
    """Get all VNet names using Azure Resource Graph.

    This is optimized for when no VNet names are provided. ARG allows a single
    query across all subscriptions, which is much faster than iterating with
    NetworkManagementClient (which requires subscription ID per call).

    Args:
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        List of all VNet names found in the subscriptions
    """
    all_vnets: list[str] = []

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Build ARG query to get all VNets across all subscriptions in a single query
        query = """
        Resources
        | where type == 'microsoft.network/virtualnetworks'
        | project name, subscriptionId, resourceGroup, id
        """

        query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=query
        )

        @retry_with_backoff(max_retries=2)
        def _query_all_vnets():
            return resource_graph_client.resources(query_request)

        response = _query_all_vnets()

        if response.data:
            for vnet_data in response.data:
                vnet_name = vnet_data.get('name')
                if vnet_name and vnet_name not in all_vnets:
                    all_vnets.append(vnet_name)

        if all_vnets:
            logging.info(
                f"Found {len(all_vnets)} VNet(s) across {len(subscription_ids)} subscription(s) (via ARG)"
            )
        else:
            logging.warning("No virtual networks found in specified subscriptions")

        return all_vnets

    except AzureError as e:
        logging.error(f"Failed to query virtual networks via ARG: {e}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error querying virtual networks via ARG: {e}")
        raise AzureError(f"Failed to query virtual networks via ARG: {e}") from e


def collect_gateway_and_firewall_resources(
    vnet_resource_ids: list[str],
    subscription_ids: list[str],
    credential: Any,
    vnet_name_to_resource_id: dict[str, str] | None = None,
    vnet_metadata: dict[str, dict] | None = None,
    ) -> dict[str, dict]:
    """Collect Virtual Network Gateways and Azure Firewalls for given VNets.

    Queries Azure Resource Graph for gateways and firewalls, then matches them
    to VNets by extracting VNet ID from their subnet configurations.

    Note on ExpressRoute:
    - virtualNetworkGateways with gatewayType="ExpressRoute" are ExpressRoute gateways
    - Additional ExpressRoute resources: Microsoft.Network/expressRouteCircuits
      and Microsoft.Network/expressRouteGateways may also exist
    - Currently only querying virtualNetworkGateways (covers both ExpressRoute and VPN)
    - TODO: Consider adding expressRouteCircuits/expressRouteGateways if needed

    Args:
        vnet_resource_ids: List of VNet resource IDs to collect gateways/firewalls for
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        Dictionary mapping VNet resource ID -> {
            'gateways': [
                {
                    'name': 'gw-name',
                    'gateway_type': 'ExpressRoute' | 'Vpn',
                    'subnet_id': '/subscriptions/.../subnets/GatewaySubnet',
                    'sku': {...},
                    'resource_id': '/subscriptions/.../virtualNetworkGateways/gw-name'
                }
            ],
            'firewalls': [
                {
                    'name': 'fw-name',
                    'subnet_id': '/subscriptions/.../subnets/AzureFirewallSubnet',
                    'resource_id': '/subscriptions/.../azureFirewalls/fw-name'
                }
            ]
        }
    """
    result: dict[str, dict] = {}

    # Initialize result dict for all VNets
    for vnet_id in vnet_resource_ids:
        result[vnet_id] = {'gateways': [], 'firewalls': []}

    logging.debug(f"Initialized gateway/firewall collection for {len(vnet_resource_ids)} VNet(s)")
    logging.debug(f"VNet resource IDs: {vnet_resource_ids[:3]}...")

    if not vnet_resource_ids:
        return result

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Query for Virtual Network Gateways
        # Note: virtualNetworkGateways can be ExpressRoute or VPN (check gatewayType property)
        gateway_query = """
        Resources
        | where type  =~ 'microsoft.network/virtualnetworkgateways'
        | project
            name,
            subscriptionId,
            resourceGroup,
            id,
            properties.gatewayType,
            properties.ipConfigurations,
            properties.sku
        """

        gateway_query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=gateway_query
        )

        @retry_with_backoff(max_retries=2)
        def _query_gateways():
            return resource_graph_client.resources(gateway_query_request)

        gateway_response = _query_gateways()

        # Log gateway query results
        if gateway_response.data:
            logging.info(f"Found {len(gateway_response.data)} virtual network gateway(s) via ARG query")
        else:
            logging.info("No virtual network gateways found via ARG query")

        # Process gateway results
        if gateway_response.data:
            for gw_data in gateway_response.data:
                try:
                    gw_id = gw_data.get('id', '')
                    gw_name = gw_data.get('name', '')

                    # Extract gatewayType from ARG response
                    # When ARG projects properties.gatewayType, it returns it as "properties_gatewayType" (with underscore)
                    # Azure returns "Vpn" or "ExpressRoute" with exact casing
                    gw_type = gw_data.get('properties_gatewayType', '')
                    if gw_type:
                        gw_type = gw_type.strip()
                    else:
                        # Log warning if gatewayType is missing - this shouldn't happen if ARG query is correct
                        logging.warning(f"Gateway {gw_name}: gatewayType not found in ARG response. Available keys: {list(gw_data.keys())}")

                    properties = gw_data.get('properties', {})

                    # ARG projects properties.ipConfigurations as "properties_ipConfigurations"
                    ip_configs = gw_data.get('properties_ipConfigurations', [])
                    if not ip_configs:
                        # Fallback to nested structure if not flattened
                        ip_configs = properties.get('ipConfigurations', [])

                    # ARG projects properties.sku as "properties_sku"
                    sku = gw_data.get('properties_sku', {})
                    if not sku:
                        # Fallback to nested structure if not flattened
                        sku = properties.get('sku', {})

                    logging.debug(f"Processing gateway: {gw_name} (type: {gw_type or 'MISSING'}, id: {gw_id})")
                    logging.debug(f"  IP configurations: {len(ip_configs) if isinstance(ip_configs, list) else 'not a list'}")

                    # Extract subnet ID from first IP configuration
                    # Structure: ipConfigurations[0].properties.subnet.id
                    # Example: /subscriptions/.../virtualNetworks/{vnet}/subnets/GatewaySubnet
                    subnet_id = None
                    if ip_configs and isinstance(ip_configs, list) and len(ip_configs) > 0:
                        first_ip_config = ip_configs[0]
                        if isinstance(first_ip_config, dict):
                            # Try nested structure first: properties.subnet.id
                            subnet_id = first_ip_config.get('properties', {}).get('subnet', {}).get('id')
                            if not subnet_id:
                                # Fallback: Try direct access if not nested
                                subnet_id = first_ip_config.get('subnet', {}).get('id')
                            # Additional fallback: Check if subnet is a direct string ID
                            if not subnet_id and 'subnet' in first_ip_config:
                                subnet_obj = first_ip_config.get('subnet')
                                if isinstance(subnet_obj, str):
                                    subnet_id = subnet_obj
                                elif isinstance(subnet_obj, dict):
                                    subnet_id = subnet_obj.get('id')

                    if subnet_id:
                        logging.debug(f"  Extracted subnet ID: {subnet_id}")
                        # Extract VNet ID from subnet ID using string split
                        # Format: /subscriptions/.../virtualNetworks/{vnet}/subnets/{subnet}
                        # Remove /subnets/{subnet} to get VNet ID
                        if '/subnets/' in subnet_id:
                            extracted_vnet_id: str | None = subnet_id.rsplit('/subnets/', 1)[0]
                        else:
                            logging.debug(f"Gateway {gw_name}: Subnet ID format unexpected: {subnet_id}")
                            extracted_vnet_id = None

                        if extracted_vnet_id:
                            # Extract VNet name from subnet ID for name-based matching
                            from gettopology.models import extract_vnet_name_from_id
                            vnet_name_from_subnet = extract_vnet_name_from_id(subnet_id)
                            logging.debug(f"Gateway {gw_name}: Extracted VNet ID from subnet: {extracted_vnet_id}")
                            logging.info(f"Gateway {gw_name}: Extracted VNet '{vnet_name_from_subnet}' from subnet")

                            # Normalize VNet ID (remove trailing slash if present, lowercase for comparison)
                            vnet_id_normalized = extracted_vnet_id.rstrip('/').lower()

                            # Try exact match first (by resource ID)
                            matched = False
                            for vnet_id_key in result.keys():
                                vnet_id_key_normalized = vnet_id_key.rstrip('/').lower()
                                if vnet_id_key_normalized == vnet_id_normalized:
                                    result[vnet_id_key]['gateways'].append({
                                        'name': gw_name,
                                        'gateway_type': gw_type,
                                        'subnet_id': subnet_id,
                                        'sku': sku,
                                        'resource_id': gw_id
                                    })
                                    # Extract VNet name for cleaner log message
                                    vnet_name_for_log = extract_vnet_name_from_id(vnet_id_key) or vnet_id_key.split('/')[-1]
                                    logging.info(f"✓ Matched gateway {gw_name} ({gw_type}) to VNet '{vnet_name_for_log}'")
                                    logging.debug(f"  Matched by resource ID (normalized from {extracted_vnet_id})")
                                    matched = True
                                    break

                            # Fallback: Try matching by VNet name
                            if not matched and vnet_name_from_subnet and vnet_name_to_resource_id:
                                if vnet_name_from_subnet in vnet_name_to_resource_id:
                                    matched_resource_id = vnet_name_to_resource_id[vnet_name_from_subnet]
                                    if matched_resource_id in result:
                                        result[matched_resource_id]['gateways'].append({
                                            'name': gw_name,
                                            'gateway_type': gw_type,
                                            'subnet_id': subnet_id,
                                            'sku': sku,
                                            'resource_id': gw_id
                                        })
                                        logging.info(f"✓ Matched gateway {gw_name} ({gw_type}) to VNet '{vnet_name_from_subnet}'")
                                        logging.debug(f"  Matched by name (extracted from subnet {subnet_id})")
                                        matched = True

                            if not matched:
                                logging.warning(f"✗ Gateway {gw_name}: VNet {extracted_vnet_id} (name: {vnet_name_from_subnet}) not found in collected VNets")
                                logging.warning(f"  Looking for: {extracted_vnet_id} (normalized: {vnet_id_normalized})")
                                logging.warning(f"  Available VNet IDs ({len(result.keys())} total):")
                                for key in list(result.keys())[:10]:
                                    logging.warning(f"    - {key} (normalized: {key.rstrip('/').lower()})")
                        else:
                            logging.debug(f"Gateway {gw_name}: Could not extract VNet ID from subnet {subnet_id}")
                    else:
                        logging.debug(f"Gateway {gw_name}: No subnet ID found in IP configurations, using fallback matching")
                        logging.debug(f"  IP config structure: {ip_configs}")
                        if ip_configs and isinstance(ip_configs, list) and len(ip_configs) > 0:
                            logging.debug(f"  First IP config keys: {list(ip_configs[0].keys()) if isinstance(ip_configs[0], dict) else 'not a dict'}")

                        # Fallback: Try to match by resource group and subscription
                        # Extract resource group and subscription from gateway ID
                        # Format: /subscriptions/{sub}/resourceGroups/{rg}/providers/.../gateways/{name}
                        if gw_id and vnet_metadata:
                            try:
                                import re
                                # Extract subscription and resource group from gateway ID
                                sub_match = re.search(r'/subscriptions/([^/]+)', gw_id)
                                rg_match = re.search(r'/resourceGroups/([^/]+)', gw_id)
                                if sub_match and rg_match:
                                    gw_subscription = sub_match.group(1)
                                    gw_resource_group = rg_match.group(1)
                                    logging.debug(f"  Gateway resource group: {gw_resource_group}, subscription: {gw_subscription}")

                                    # Try to match to VNets in the same resource group and subscription
                                    for vnet_id_key, vnet_info in vnet_metadata.items():
                                        if (vnet_info.get('resource_group') == gw_resource_group and
                                            vnet_info.get('subscription_id') == gw_subscription):
                                            # Found a VNet in the same resource group and subscription
                                            # This is a reasonable match (gateways are typically in same RG as VNet)
                                            from gettopology.models import extract_vnet_name_from_id
                                            result[vnet_id_key]['gateways'].append({
                                                'name': gw_name,
                                                'gateway_type': gw_type,
                                                'subnet_id': None,  # Couldn't extract from IP configs
                                                'sku': sku,
                                                'resource_id': gw_id
                                            })
                                            vnet_name_fallback = extract_vnet_name_from_id(vnet_id_key) or vnet_id_key.split('/')[-1]
                                            logging.info(f"✓ Matched gateway {gw_name} ({gw_type}) to VNet '{vnet_name_fallback}'")
                                            logging.debug(f"  Matched by resource group '{gw_resource_group}' and subscription (fallback match)")
                                            break
                            except Exception as e:
                                logging.debug(f"  Failed to match gateway by resource group: {e}")
                except Exception as e:
                    logging.debug(f"Failed to process gateway {gw_data.get('name', 'unknown')}: {e}")
                    continue

        # Query for Azure Firewalls
        firewall_query = """
        Resources
        | where type =~ 'microsoft.network/azurefirewalls'
        | project
            name,
            subscriptionId,
            resourceGroup,
            id,
            properties.ipConfigurations[0].properties.subnet.id,
            properties.sku
        """

        firewall_query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=firewall_query
        )

        @retry_with_backoff(max_retries=2)
        def _query_firewalls():
            return resource_graph_client.resources(firewall_query_request)

        firewall_response = _query_firewalls()

        # Process firewall results
        if firewall_response.data:
            for fw_data in firewall_response.data:
                try:
                    fw_id = fw_data.get('id', '')
                    fw_name = fw_data.get('name', '')
                    # Try to get subnet ID directly from projected field (ARG may flatten nested paths)
                    subnet_id = fw_data.get('properties.ipConfigurations[0].properties.subnet.id') or fw_data.get('properties_ipConfigurations_0_properties_subnet_id')
                    sku = fw_data.get('properties', {}).get('sku', {}) if isinstance(fw_data.get('properties'), dict) else {}

                    # Fallback: Extract from nested properties structure if direct access didn't work
                    if not subnet_id:
                        fw_properties = fw_data.get('properties', {})
                        if isinstance(fw_properties, dict):
                            ip_configs = fw_properties.get('ipConfigurations', [])
                            if ip_configs and isinstance(ip_configs, list) and len(ip_configs) > 0:
                                first_ip_config = ip_configs[0]
                                if isinstance(first_ip_config, dict):
                                    ip_config_props = first_ip_config.get('properties', {})
                                    if ip_config_props:
                                        subnet_obj = ip_config_props.get('subnet')
                                        if isinstance(subnet_obj, dict):
                                            subnet_id = subnet_obj.get('id')
                                        elif isinstance(subnet_obj, str):
                                            subnet_id = subnet_obj

                    if not subnet_id:
                        logging.warning(f"Firewall {fw_name}: No subnet ID found")
                        continue

                    # Extract VNet ID from subnet ID
                    if '/subnets/' in subnet_id:
                        vnet_id = subnet_id.rsplit('/subnets/', 1)[0]
                    else:
                        logging.warning(f"Firewall {fw_name}: Subnet ID format unexpected: {subnet_id}")
                        continue

                    # Extract VNet name from subnet ID for name-based matching
                    from gettopology.models import extract_vnet_name_from_id
                    vnet_name_from_subnet = extract_vnet_name_from_id(subnet_id)

                    # Normalize VNet ID (remove trailing slash if present, lowercase for comparison)
                    vnet_id_normalized = vnet_id.rstrip('/').lower()

                    # Try exact match first (by resource ID)
                    matched = False
                    for vnet_id_key in result.keys():
                        vnet_id_key_normalized = vnet_id_key.rstrip('/').lower()
                        if vnet_id_key_normalized == vnet_id_normalized:
                            result[vnet_id_key]['firewalls'].append({
                                'name': fw_name,
                                'subnet_id': subnet_id,
                                'sku': sku,
                                'resource_id': fw_id
                            })
                            matched = True
                            break

                    # Fallback: Try matching by VNet name
                    if not matched and vnet_name_from_subnet and vnet_name_to_resource_id:
                        if vnet_name_from_subnet in vnet_name_to_resource_id:
                            matched_resource_id = vnet_name_to_resource_id[vnet_name_from_subnet]
                            if matched_resource_id in result:
                                result[matched_resource_id]['firewalls'].append({
                                    'name': fw_name,
                                    'subnet_id': subnet_id,
                                    'sku': sku,
                                    'resource_id': fw_id
                                })
                                matched = True

                    if not matched:
                        logging.warning(f"Firewall {fw_name}: VNet {vnet_id} (name: {vnet_name_from_subnet}) not found in collected VNets")
                except Exception as e:
                    logging.error(f"Failed to process firewall {fw_data.get('name', 'unknown')}: {e}", exc_info=True)
                    continue

        return result

    except AzureError as e:
        logging.warning(f"Failed to collect gateway/firewall resources via ARG: {e}")
        # Return empty result dict (backward compatibility)
        return result
    except Exception as e:
        logging.warning(f"Unexpected error collecting gateway/firewall resources: {e}")
        # Return empty result dict (backward compatibility)
        return result


def query_route_table_routes(
    route_table_ids: list[str],
    subscription_ids: list[str],
    credential: Any,
    ) -> dict[str, list[dict]]:
    """Query route table routes from Azure Resource Graph.

    Args:
        route_table_ids: List of route table resource IDs
        subscription_ids: List of subscription IDs
        credential: Azure credential object

    Returns:
        Dictionary mapping route table ID -> list of route dicts
        Each route dict contains: name, addressPrefix, nextHopType, nextHopIpAddress
    """
    result: dict[str, list[dict]] = {}

    for rt_id in route_table_ids:
        result[rt_id] = []

    if not route_table_ids:
        return result

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Extract route table names from IDs
        rt_names = []
        for rt_id in route_table_ids:
            parts = rt_id.split('/routeTables/')
            if len(parts) == 2:
                rt_names.append(parts[1])

        if not rt_names:
            return result

        name_conditions = " or ".join([f"name == '{name}'" for name in rt_names])
        query = f"""
        Resources
        | where type == "microsoft.network/routetables"
        | where {name_conditions}
        | project name, id, properties.routes
        """

        query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=query
        )

        @retry_with_backoff(max_retries=2)
        def _query_route_tables():
            return resource_graph_client.resources(query_request)

        response = _query_route_tables()

        if response.data:
            for rt_data in response.data:
                rt_id = rt_data.get('id', '')
                if rt_id not in route_table_ids:
                    continue

                routes = rt_data.get('properties_routes', [])
                if not routes:
                    props = rt_data.get('properties', {})
                    routes = props.get('routes', [])

                route_list = []
                if isinstance(routes, list):
                    for route in routes:
                        if isinstance(route, dict):
                            route_props = route.get('properties', {})
                            if not route_props:
                                route_props = route

                            route_dict = {
                                'name': route.get('name', ''),
                                'addressPrefix': route_props.get('addressPrefix', ''),
                                'nextHopType': route_props.get('nextHopType', ''),
                                'nextHopIpAddress': route_props.get('nextHopIpAddress', '') or '—',
                            }
                            route_list.append(route_dict)

                result[rt_id] = route_list

        return result
    except Exception as e:
        logging.warning(f"Failed to query route table routes: {e}")
        return result


def collect_topology(
    vnet_names: list[str],
    subscription_ids: list[str],
    credential: Any,
    ) -> TopologyModel:
    """Collect complete topology data for virtual networks.

    Queries Azure Resource Graph to get full VNet details including subnets,
    peerings, and all properties, then parses them into TopologyModel.

    Args:
        vnet_names: List of VNet names to collect topology for
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        TopologyModel containing all virtual networks with their complete topology

    Raises:
        ValueError: If credential is required but not provided
        AzureError: If Azure API calls fail
    """
    if credential is None:
        raise ValueError("credential is required")

    if not subscription_ids:
        raise ValueError("subscription_ids cannot be empty")

    if not vnet_names:
        logging.warning("No VNet names provided, returning empty topology")
        return TopologyModel(
            virtual_networks=[],
            collected_at=datetime.utcnow().isoformat()
        )

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Build ARG query to get full VNet details
        # Query all properties without filtering to get complete nested structure
        name_conditions = " or ".join([f"name == '{name}'" for name in vnet_names])
        query = f"""
        Resources
        | where type == 'microsoft.network/virtualnetworks'
        | where {name_conditions}
        """

        query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=query
        )

        logging.info(f"Collecting topology data for {len(vnet_names)} VNet(s) via ARG...")

        @retry_with_backoff(max_retries=2)
        def _query_topology():
            return resource_graph_client.resources(query_request)

        response = _query_topology()

        # Get subscription names for all unique subscription IDs
        unique_subscription_ids = list(set(subscription_ids))
        subscription_names_map = get_subscription_names(unique_subscription_ids, credential)

        virtual_networks: list[VirtualNetworkModel] = []

        if response.data:
            for vnet_data in response.data:
                try:
                    # Add subscription_name to vnet_data before parsing
                    vnet_subscription_id = vnet_data.get('subscriptionId', '')
                    if vnet_subscription_id in subscription_names_map:
                        vnet_data['subscriptionName'] = subscription_names_map[vnet_subscription_id]

                    # Parse ARG response into VirtualNetworkModel
                    # The model's @model_validator will handle nested structure extraction
                    vnet = VirtualNetworkModel(**vnet_data)
                    virtual_networks.append(vnet)
                    logging.debug(f"Collected topology for VNet: {vnet.name}")
                except Exception as e:
                    logging.warning(f"Failed to parse VNet data for {vnet_data.get('name', 'unknown')}: {e}")
                    continue

        # Collect gateway and firewall resources for VNets
        try:
            vnet_resource_ids = [vnet.resource_id for vnet in virtual_networks if vnet.resource_id]
            # Create mapping of VNet name to resource_id for name-based matching
            vnet_name_to_resource_id = {vnet.name: vnet.resource_id for vnet in virtual_networks if vnet.resource_id}
            # Create mapping of VNet resource_id to resource_group and subscription for fallback matching
            vnet_metadata = {}
            for vnet in virtual_networks:
                if vnet.resource_id:
                    vnet_metadata[vnet.resource_id] = {
                        'resource_group': vnet.resource_group_name,
                        'subscription_id': vnet.subscription_id,
                        'name': vnet.name
                    }
            logging.debug(f"Collecting gateway/firewall resources for {len(vnet_resource_ids)} VNet(s) with resource IDs")
            if vnet_resource_ids:
                gateway_firewall_data = collect_gateway_and_firewall_resources(
                    vnet_resource_ids,
                    subscription_ids,
                    credential,
                    vnet_name_to_resource_id,  # Pass name mapping for fallback matching
                    vnet_metadata  # Pass VNet metadata for resource group matching
                )

                # Attach gateway/firewall data to each VNet
                # Use model_copy() to update fields (Pydantic v2 safe way)
                updated_vnets = []
                for vnet in virtual_networks:
                    if vnet.resource_id:
                        # Try exact match first
                        if vnet.resource_id in gateway_firewall_data:
                            fw_data = gateway_firewall_data[vnet.resource_id].get('firewalls', [])
                            gw_data = gateway_firewall_data[vnet.resource_id].get('gateways', [])
                            updated_vnet = vnet.model_copy(update={
                                'virtual_network_gateways': gw_data,
                                'azure_firewalls': fw_data
                            })
                            updated_vnets.append(updated_vnet)
                        else:
                            # Try normalized match (case-insensitive, trailing slash)
                            vnet_id_normalized = vnet.resource_id.rstrip('/').lower()
                            matched = False
                            for key in gateway_firewall_data.keys():
                                key_normalized = key.rstrip('/').lower()
                                if key_normalized == vnet_id_normalized:
                                    fw_data = gateway_firewall_data[key].get('firewalls', [])
                                    gw_data = gateway_firewall_data[key].get('gateways', [])
                                    updated_vnet = vnet.model_copy(update={
                                        'virtual_network_gateways': gw_data,
                                        'azure_firewalls': fw_data
                                    })
                                    updated_vnets.append(updated_vnet)
                                    matched = True
                                    break
                            if not matched:
                                updated_vnets.append(vnet)
                    else:
                        updated_vnets.append(vnet)
                virtual_networks = updated_vnets
        except Exception as e:
            # If gateway/firewall collection fails, log but don't fail entire collection
            # VNets will still work with fallback subnet-based detection
            logging.warning(f"Failed to collect gateway/firewall resources: {e}")

        logging.info(f"Successfully collected topology for {len(virtual_networks)} VNet(s)")

        return TopologyModel(
            virtual_networks=virtual_networks,
            collected_at=datetime.utcnow().isoformat()
        )

    except AzureError as e:
        logging.error(f"Failed to collect topology via ARG: {e}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error collecting topology: {e}")
        raise AzureError(f"Failed to collect topology: {e}") from e


def collect_connections(
    subscription_ids: list[str],
    credential: Any,
    ) -> list[ConnectionModel]:
    """Collect all Connections (VPN and ExpressRoute) using Azure Resource Graph.

    Args:
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        List of ConnectionModel objects
    """
    connections: list[ConnectionModel] = []

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Query for Connections
        # Note: ARG flattens nested properties with underscores
        # properties.virtualNetworkGateway1.id becomes properties_virtualNetworkGateway1_id
        # Some fields may be missing depending on connection type
        connection_query = """
        Resources
        | where type =~ 'microsoft.network/connections'
        | project
            name,
            subscriptionId,
            resourceGroup,
            location,
            id,
            properties.connectionType,
            properties.connectionStatus,
            properties.connectionProtocol,
            properties.connectionMode,
            properties.virtualNetworkGateway1,
            properties.localNetworkGateway2,
            properties.peer,
            properties.routingWeight,
            properties.sharedKey,
            properties.authenticationType,
            properties.enableBgp,
            properties.usePolicyBasedTrafficSelectors,
            properties.ipsecPolicies,
            properties.dpdTimeoutSeconds,
            properties.ingressBytesTransferred,
            properties.egressBytesTransferred
        """

        query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=connection_query
        )

        @retry_with_backoff(max_retries=2)
        def _query_connections():
            return resource_graph_client.resources(query_request)

        response = _query_connections()

        if response.data:
            logging.info(f"Found {len(response.data)} connection(s) via ARG query")
            for conn_data in response.data:
                try:
                    # ARG returns properties with underscores (e.g., properties_connectionType)
                    # For nested objects like virtualNetworkGateway1, we need to extract the id field
                    vnet_gw1 = conn_data.get('properties_virtualNetworkGateway1', {})
                    vnet_gateway_id = vnet_gw1.get('id') if isinstance(vnet_gw1, dict) else None
                    
                    lng_gw2 = conn_data.get('properties_localNetworkGateway2', {})
                    local_gateway_id = lng_gw2.get('id') if isinstance(lng_gw2, dict) else None
                    
                    peer_obj = conn_data.get('properties_peer', {})
                    peer_id = peer_obj.get('id') if isinstance(peer_obj, dict) else None

                    # IPsec policies may be an array or None
                    ipsec_policies = conn_data.get('properties_ipsecPolicies', [])
                    if not isinstance(ipsec_policies, list):
                        ipsec_policies = []

                    connection = ConnectionModel(
                        name=conn_data.get('name', ''),
                        resource_id=conn_data.get('id', ''),
                        subscription_id=conn_data.get('subscriptionId', ''),
                        resource_group_name=conn_data.get('resourceGroup', ''),
                        location=conn_data.get('location', ''),
                        connection_type=conn_data.get('properties_connectionType', 'IPsec'),
                        connection_status=conn_data.get('properties_connectionStatus'),
                        connection_protocol=conn_data.get('properties_connectionProtocol'),
                        connection_mode=conn_data.get('properties_connectionMode'),
                        virtual_network_gateway_id=vnet_gateway_id,
                        local_network_gateway_id=local_gateway_id,
                        peer_id=peer_id,
                        routing_weight=conn_data.get('properties_routingWeight'),
                        shared_key=conn_data.get('properties_sharedKey'),  # May be None for security
                        authentication_type=conn_data.get('properties_authenticationType'),
                        enable_bgp=conn_data.get('properties_enableBgp', False),
                        use_policy_based_traffic_selectors=conn_data.get('properties_usePolicyBasedTrafficSelectors', False),
                        ipsec_policies=ipsec_policies,
                        dpd_timeout_seconds=conn_data.get('properties_dpdTimeoutSeconds'),
                        ingress_bytes_transferred=conn_data.get('properties_ingressBytesTransferred'),
                        egress_bytes_transferred=conn_data.get('properties_egressBytesTransferred'),
                    )
                    connections.append(connection)
                except Exception as e:
                    logging.warning(f"Failed to parse connection data for {conn_data.get('name', 'unknown')}: {e}")
                    logging.debug(f"Connection data keys: {list(conn_data.keys())}")
                    continue
        else:
            logging.info("No connections found via ARG query")

        return connections

    except AzureError as e:
        logging.error(f"Failed to query connections via ARG: {e}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error collecting connections: {e}")
        raise AzureError(f"Failed to collect connections: {e}") from e


def collect_local_network_gateways(
    subscription_ids: list[str],
    credential: Any,
    ) -> list[LocalNetworkGatewayModel]:
    """Collect all Local Network Gateways using Azure Resource Graph.

    Args:
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        List of LocalNetworkGatewayModel objects
    """
    local_gateways: list[LocalNetworkGatewayModel] = []

    try:
        resource_graph_client = ResourceGraphClient(credential)

        # Query for Local Network Gateways
        lng_query = """
        Resources
        | where type == 'microsoft.network/localnetworkgateways'
        | project
            name,
            subscriptionId,
            resourceGroup,
            location,
            id,
            properties.provisioningState,
            properties.resourceGuid,
            properties.gatewayIpAddress,
            properties.fqdn,
            properties.localNetworkAddressSpace.addressPrefixes,
            properties.bgpSettings
        """

        query_request = QueryRequest(
            subscriptions=subscription_ids,
            query=lng_query
        )

        @retry_with_backoff(max_retries=2)
        def _query_local_gateways():
            return resource_graph_client.resources(query_request)

        response = _query_local_gateways()

        if response.data:
            logging.info(f"Found {len(response.data)} local network gateway(s) via ARG query")
            for lng_data in response.data:
                try:
                    # ARG returns properties with underscores
                    address_prefixes = lng_data.get('properties_localNetworkAddressSpace_addressPrefixes', [])
                    if not isinstance(address_prefixes, list):
                        address_prefixes = []

                    # BGP settings may be an object or None
                    bgp_settings = lng_data.get('properties_bgpSettings')
                    if not isinstance(bgp_settings, dict):
                        bgp_settings = None

                    # FQDN may be empty string or None
                    fqdn = lng_data.get('properties_fqdn')
                    if fqdn == '':
                        fqdn = None

                    local_gateway = LocalNetworkGatewayModel(
                        name=lng_data.get('name', ''),
                        resource_id=lng_data.get('id', ''),
                        subscription_id=lng_data.get('subscriptionId', ''),
                        resource_group_name=lng_data.get('resourceGroup', ''),
                        location=lng_data.get('location', ''),
                        provisioning_state=lng_data.get('properties_provisioningState'),
                        resource_guid=lng_data.get('properties_resourceGuid'),
                        gateway_ip_address=lng_data.get('properties_gatewayIpAddress'),
                        fqdn=fqdn,
                        address_prefixes=address_prefixes,
                        bgp_settings=bgp_settings,
                    )
                    local_gateways.append(local_gateway)
                except Exception as e:
                    logging.warning(f"Failed to parse local network gateway data for {lng_data.get('name', 'unknown')}: {e}")
                    continue
        else:
            logging.info("No local network gateways found via ARG query")

        return local_gateways

    except AzureError as e:
        logging.error(f"Failed to query local network gateways via ARG: {e}")
        raise
    except Exception as e:
        logging.error(f"Unexpected error collecting local network gateways: {e}")
        raise AzureError(f"Failed to collect local network gateways: {e}") from e


def collect_hybrid_connectivity(
    topology: TopologyModel,
    subscription_ids: list[str],
    credential: Any,
    ) -> dict[str, Any]:
    """Collect and correlate hybrid connectivity data (Connections and Local Network Gateways).

    Correlates connections and local network gateways with existing VNets and Gateways from topology.

    Args:
        topology: TopologyModel containing VNets and gateways
        subscription_ids: List of subscription IDs to search in
        credential: Azure credential object

    Returns:
        Dictionary with correlated hybrid connectivity data:
        {
            'connections': [ConnectionModel, ...],
            'local_network_gateways': [LocalNetworkGatewayModel, ...],
            'vnet_to_connections': {vnet_name: [ConnectionModel, ...]},
            'gateway_to_connections': {gateway_resource_id: [ConnectionModel, ...]},
            'connection_to_local_gateway': {connection_resource_id: LocalNetworkGatewayModel},
            'connection_to_peer_gateway': {connection_resource_id: VirtualNetworkGateway dict}
        }
    """
    logging.info("Collecting hybrid connectivity data (Connections and Local Network Gateways)...")

    # Collect connections and local network gateways
    connections = collect_connections(subscription_ids, credential)
    local_network_gateways = collect_local_network_gateways(subscription_ids, credential)

    # Helper function to normalize resource IDs for matching (case-insensitive, strip trailing slashes)
    def normalize_resource_id(resource_id: str | None) -> str | None:
        if not resource_id:
            return None
        return resource_id.rstrip('/').lower()

    # Create mappings for correlation (with normalized IDs for better matching)
    local_gateway_by_id: dict[str, LocalNetworkGatewayModel] = {}
    local_gateway_by_normalized_id: dict[str, LocalNetworkGatewayModel] = {}
    for lng in local_network_gateways:
        if lng.resource_id:
            local_gateway_by_id[lng.resource_id] = lng
            normalized = normalize_resource_id(lng.resource_id)
            if normalized:
                local_gateway_by_normalized_id[normalized] = lng

    vnet_by_resource_id = {vnet.resource_id: vnet for vnet in topology.virtual_networks if vnet.resource_id}
    gateway_by_resource_id: dict[str, dict] = {}
    gateway_by_normalized_id: dict[str, dict] = {}
    gateway_to_vnet: dict[str, str] = {}  # Map gateway resource_id -> vnet_name

    # Build gateway mapping from VNets
    for vnet in topology.virtual_networks:
        for gateway in vnet.virtual_network_gateways:
            gw_resource_id = gateway.get('resource_id')
            if gw_resource_id:
                gateway_by_resource_id[gw_resource_id] = gateway
                normalized = normalize_resource_id(gw_resource_id)
                if normalized:
                    gateway_by_normalized_id[normalized] = gateway
                    gateway_to_vnet[gw_resource_id] = vnet.name
                    # Also map normalized ID to VNet
                    if normalized:
                        gateway_to_vnet[normalized] = vnet.name

    # Correlate connections with VNets and Gateways
    vnet_to_connections: dict[str, list[ConnectionModel]] = {}
    gateway_to_connections: dict[str, list[ConnectionModel]] = {}
    connection_to_local_gateway: dict[str, LocalNetworkGatewayModel] = {}
    connection_to_peer_gateway: dict[str, dict] = {}

    for connection in connections:
        # Find VNet from gateway (try exact match first, then normalized)
        vnet_name = None
        if connection.virtual_network_gateway_id:
            # Try exact match first
            vnet_name = gateway_to_vnet.get(connection.virtual_network_gateway_id)
            
            # If not found, try normalized match
            if not vnet_name:
                normalized_gw_id = normalize_resource_id(connection.virtual_network_gateway_id)
                if normalized_gw_id:
                    vnet_name = gateway_to_vnet.get(normalized_gw_id)
                    # Also update the gateway mapping if we found it via normalized ID
                    if normalized_gw_id in gateway_by_normalized_id:
                        gateway_by_resource_id[connection.virtual_network_gateway_id] = gateway_by_normalized_id[normalized_gw_id]

        # Map connection to VNet
        if vnet_name:
            if vnet_name not in vnet_to_connections:
                vnet_to_connections[vnet_name] = []
            vnet_to_connections[vnet_name].append(connection)

        # Map connection to gateway (exact or normalized match)
        if connection.virtual_network_gateway_id:
            gateway = gateway_by_resource_id.get(connection.virtual_network_gateway_id)
            if not gateway:
                normalized_gw_id = normalize_resource_id(connection.virtual_network_gateway_id)
                if normalized_gw_id:
                    gateway = gateway_by_normalized_id.get(normalized_gw_id)
            
            if gateway:
                # Use normalized ID as key for consistency
                normalized_key = normalize_resource_id(connection.virtual_network_gateway_id) or connection.virtual_network_gateway_id
                if normalized_key not in gateway_to_connections:
                    gateway_to_connections[normalized_key] = []
                gateway_to_connections[normalized_key].append(connection)

        # Map connection to local network gateway (for S2S VPN)
        if connection.local_network_gateway_id:
            # Try exact match first
            local_gateway = local_gateway_by_id.get(connection.local_network_gateway_id)
            # If not found, try normalized match
            if not local_gateway:
                normalized_lng_id = normalize_resource_id(connection.local_network_gateway_id)
                if normalized_lng_id:
                    local_gateway = local_gateway_by_normalized_id.get(normalized_lng_id)
            
            if local_gateway:
                connection_to_local_gateway[connection.resource_id] = local_gateway

        # Map connection to peer gateway (for VNet-to-VNet) or ExpressRoute Circuit
        if connection.peer_id:
            # Check if it's an ExpressRoute Circuit (not a gateway)
            if 'expressRouteCircuits' in connection.peer_id.lower():
                # ExpressRoute Circuit - no gateway correlation needed
                pass
            else:
                # Try to find peer gateway
                peer_gateway = gateway_by_resource_id.get(connection.peer_id)
                if not peer_gateway:
                    normalized_peer_id = normalize_resource_id(connection.peer_id)
                    if normalized_peer_id:
                        peer_gateway = gateway_by_normalized_id.get(normalized_peer_id)
                
                if peer_gateway:
                    connection_to_peer_gateway[connection.resource_id] = peer_gateway

    # Calculate correlation statistics
    correlated_connections = sum(len(conns) for conns in vnet_to_connections.values())
    uncorrelated_connections = len(connections) - correlated_connections
    connections_with_lng = len(connection_to_local_gateway)
    connections_with_peer = len(connection_to_peer_gateway)

    result = {
        'connections': connections,
        'local_network_gateways': local_network_gateways,
        'vnet_to_connections': vnet_to_connections,
        'gateway_to_connections': gateway_to_connections,
        'connection_to_local_gateway': connection_to_local_gateway,
        'connection_to_peer_gateway': connection_to_peer_gateway,
    }

    logging.info(
        f"Hybrid connectivity collected: {len(connections)} connection(s), "
        f"{len(local_network_gateways)} local network gateway(s)"
    )
    logging.debug(
        f"Correlation stats: {correlated_connections} connection(s) correlated with VNets, "
        f"{uncorrelated_connections} uncorrelated, {connections_with_lng} with Local Network Gateway, "
        f"{connections_with_peer} with peer gateway/circuit"
    )

    # Log uncorrelated connections for debugging
    if uncorrelated_connections > 0:
        uncorrelated = [
            conn for conn in connections
            if not any(conn.resource_id in [c.resource_id for c in conns] for conns in vnet_to_connections.values())
        ]
        for conn in uncorrelated:
            logging.debug(
                f"Uncorrelated connection: {conn.name} "
                f"(Gateway ID: {conn.virtual_network_gateway_id})"
            )

    return result



"""Main page generation for diagram.

Creates the main page with hubs, spokes, and external VNets.
"""

from xml.etree.ElementTree import Element

from gettopology.diagram.config import get_config
from gettopology.diagram.elements.vnet_elements import (
    calculate_vnet_height,
    create_external_vnet_group,
    create_vnet_group,
)
from gettopology.diagram.layout import (
    Zone,
    calculate_external_zone_positions,
    calculate_spoke_positions,
)
from gettopology.diagram.legend import create_legend
from gettopology.models import TopologyModel

# Get global config instance
_config = get_config()

# Constants
CANVAS_PADDING = _config.canvas_padding
VNET_WIDTH = _config.vnet_width
VNET_MIN_HEIGHT = _config.vnet_min_height
EXTERNAL_SPACING_Y = _config.external_spacing_y
SPACING_BELOW_HUB = _config.spacing_below_hub
SPACING_BETWEEN_SPOKES = _config.spacing_between_spokes


def create_main_page(
    root: Element,
    topology: TopologyModel,
    zones: list[Zone],
    primary_subscription_id: str | None,
    hubless_spoke_names: set[str],
    stubs: list,
    hub_names: set[str],
    vnet_name_to_tenant_id: dict[str, str],
    hubless_group_map: dict[str, int],
    hubless_centers: dict[int, str],
    cell_id_counter: int,
    legend_x: float,
    create_peering_edge_func,
) -> tuple[
    int,  # cell_id_counter
    dict[str, int],  # vnet_name_to_main_id
    dict[str, tuple[float, float, float]],  # hub_name_to_position
    dict[str, tuple[float, float, float]],  # regular_spoke_positions
    dict[str, tuple[float, float, float]],  # hubless_spoke_positions
    dict[str, tuple[float, float, float]],  # all_vnet_positions
    list,  # edges_to_create
]:
    """Create main page with hubs, spokes, and external VNets.

    Returns updated state dictionaries and cell_id_counter.
    """
    # Initialize dictionaries
    vnet_name_to_main_id: dict[str, int] = {}
    hub_name_to_position: dict[str, tuple[float, float, float]] = {}
    regular_spoke_positions: dict[str, tuple[float, float, float]] = {}
    hubless_spoke_positions: dict[str, tuple[float, float, float]] = {}
    all_vnet_positions: dict[str, tuple[float, float, float]] = {}
    edges_to_create: list = []
    created_edges: set[tuple[int, int]] = set()

    def _add_edge_if_not_exists(
        source_id: int,
        target_id: int,
        is_local: bool,
        source_name: str,
        target_name: str,
        source_x: float | None = None,
        target_x: float | None = None,
        source_y: float | None = None,
        target_y: float | None = None,
        source_height: float | None = None,
        target_height: float | None = None,
    ) -> None:
        """Helper to add edge to list only if it doesn't already exist."""
        sorted_ids = sorted([source_id, target_id])
        edge_key: tuple[int, int] = (sorted_ids[0], sorted_ids[1])
        if edge_key not in created_edges:
            is_hubless_center = False
            if hubless_centers:
                is_hubless_center = source_name in hubless_centers.values()
            source_is_hub = (source_name in hub_names and source_name not in hubless_spoke_names) or is_hubless_center
            target_is_hub = target_name in hub_names and target_name not in hubless_spoke_names
            source_tenant_id = vnet_name_to_tenant_id.get(source_name)
            target_tenant_id = vnet_name_to_tenant_id.get(target_name)
            is_cross_tenant = (source_tenant_id is not None and target_tenant_id is not None and source_tenant_id != target_tenant_id) or (target_name not in vnet_name_to_tenant_id)
            edges_to_create.append((source_id, target_id, is_local, source_is_hub, target_is_hub, source_x, target_x, source_y, target_y, source_height, target_height, source_name, target_name, is_cross_tenant))
            created_edges.add(edge_key)

    # PHASE 1: Create all VNets first (for proper z-ordering, edges will be created after)
    # 1. Layout hub-spoke groups using calculated zones
    zone_bottoms: list[float] = []
    for zone in zones:
        hub_x = zone.zone_x
        hub_y_adjusted = zone.zone_y
        hub_height = calculate_vnet_height(zone.hub)

        # Draw hub
        group_id, hub_main_id, cell_id_counter = create_vnet_group(
            root, cell_id_counter, zone.hub, hub_x, hub_y_adjusted,
            is_hub=True, primary_subscription_id=primary_subscription_id
        )
        vnet_name_to_main_id[zone.hub.name] = hub_main_id
        hub_name_to_position[zone.hub.name] = (hub_x, hub_y_adjusted, hub_height)
        all_vnet_positions[zone.hub.name] = (hub_x, hub_y_adjusted, hub_height)

        # Calculate spoke positions using layout module
        left_spoke_positions = calculate_spoke_positions(zone, is_left=True)
        right_spoke_positions = calculate_spoke_positions(zone, is_left=False)

        # Create left spokes
        for spoke, x_position, y_position in left_spoke_positions:
            spoke_height = calculate_vnet_height(spoke)
            group_id, spoke_main_id, cell_id_counter = create_vnet_group(
                root, cell_id_counter, spoke, x_position, y_position,
                is_hub=False, primary_subscription_id=primary_subscription_id,
                hub_subscription_id=zone.hub.subscription_id
            )
            vnet_name_to_main_id[spoke.name] = spoke_main_id
            regular_spoke_positions[spoke.name] = (x_position, y_position, spoke_height)
            all_vnet_positions[spoke.name] = (x_position, y_position, spoke_height)

            # Store edge information
            peering_index = spoke.peering_names.index(zone.hub.name) if zone.hub.name in spoke.peering_names else -1
            is_local = spoke.peering_local_flags[peering_index] if peering_index >= 0 and peering_index < len(spoke.peering_local_flags) else True
            _add_edge_if_not_exists(
                hub_main_id, spoke_main_id, is_local, zone.hub.name, spoke.name,
                hub_x, x_position, hub_y_adjusted, y_position, hub_height, spoke_height
            )

        # Create right spokes
        for spoke, x_position, y_position in right_spoke_positions:
            spoke_height = calculate_vnet_height(spoke)
            group_id, spoke_main_id, cell_id_counter = create_vnet_group(
                root, cell_id_counter, spoke, x_position, y_position,
                is_hub=False, primary_subscription_id=primary_subscription_id,
                hub_subscription_id=zone.hub.subscription_id
            )
            vnet_name_to_main_id[spoke.name] = spoke_main_id
            regular_spoke_positions[spoke.name] = (x_position, y_position, spoke_height)
            all_vnet_positions[spoke.name] = (x_position, y_position, spoke_height)

            # Store edge information
            peering_index = spoke.peering_names.index(zone.hub.name) if zone.hub.name in spoke.peering_names else -1
            is_local = spoke.peering_local_flags[peering_index] if peering_index >= 0 and peering_index < len(spoke.peering_local_flags) else True
            _add_edge_if_not_exists(
                hub_main_id, spoke_main_id, is_local, zone.hub.name, spoke.name,
                hub_x, x_position, hub_y_adjusted, y_position, hub_height, spoke_height
            )

        # Track zone bottom
        max_spoke_count = max(len(left_spoke_positions), len(right_spoke_positions))
        if max_spoke_count > 0:
            total_spoke_height = sum(calculate_vnet_height(spoke) for spoke, _, _ in (left_spoke_positions + right_spoke_positions))
            zone_bottom = hub_y_adjusted + hub_height + total_spoke_height + (max_spoke_count * SPACING_BETWEEN_SPOKES) + SPACING_BELOW_HUB
        else:
            zone_bottom = hub_y_adjusted + hub_height + SPACING_BELOW_HUB
        zone_bottoms.append(zone_bottom)

    # 1b. Store edges between hubs if they peer with each other
    for i, zone1 in enumerate(zones):
        hub1_main_id = vnet_name_to_main_id[zone1.hub.name]
        hub1_x, hub1_y, hub1_height = hub_name_to_position[zone1.hub.name]
        hub1_center_y = hub1_y + (hub1_height / 2)
        for j, zone2 in enumerate(zones):
            if i < j and zone2.hub.name in zone1.hub.peering_names:
                hub2_main_id = vnet_name_to_main_id[zone2.hub.name]
                hub2_x, hub2_y, hub2_height = hub_name_to_position[zone2.hub.name]
                hub2_center_y = hub2_y + (hub2_height / 2)
                peering_index = zone1.hub.peering_names.index(zone2.hub.name)
                is_local = zone1.hub.peering_local_flags[peering_index] if peering_index < len(zone1.hub.peering_local_flags) else True
                # Pass coordinates so intermediate hub detection can work
                _add_edge_if_not_exists(hub1_main_id, hub2_main_id, is_local, zone1.hub.name, zone2.hub.name, hub1_x, hub2_x, hub1_center_y, hub2_center_y, hub1_height, hub2_height)

    # PHASE 1.5: Create external VNets using zone-based layout
    # Collect all external peerings
    external_vnets: dict[str, list[tuple[str, str]]] = {}
    for vnet in topology.virtual_networks:
        if vnet.name in hubless_spoke_names:
            continue
        if vnet.name in {s.name for s in stubs}:
            continue

        for i, peer_name in enumerate(vnet.peering_names):
            if peer_name in vnet_name_to_main_id:
                continue
            if peer_name in hubless_spoke_names:
                continue
            if peer_name in {s.name for s in stubs}:
                continue
            peering_resource_id = vnet.peering_resource_ids[i] if i < len(vnet.peering_resource_ids) else ""
            if peer_name not in external_vnets:
                external_vnets[peer_name] = []
            external_vnets[peer_name].append((peering_resource_id, vnet.name))

    # Filter external VNets for main page
    external_vnets_for_main_page: dict[str, list[tuple[str, str]]] = {}
    for external_vnet_name, peerings in external_vnets.items():
        peers_with_non_hubless = any(
            source_vnet_name not in hubless_spoke_names
            for _, source_vnet_name in peerings
        )
        if peers_with_non_hubless:
            external_vnets_for_main_page[external_vnet_name] = peerings

    # Calculate external zones for main page
    external_zones = calculate_external_zone_positions(
        external_vnets_for_main_page,
        zones,
        None,
        regular_spoke_positions,
        {},
        hub_name_to_position,
    )

    # Create legend
    legend_y = CANVAS_PADDING
    cell_id_counter, actual_legend_height = create_legend(root, cell_id_counter, legend_x, legend_y)

    # Create external VNets from zones
    zones_by_peer: dict[str, list] = {}
    for external_zone in external_zones:
        if external_zone.peer_vnet_name not in zones_by_peer:
            zones_by_peer[external_zone.peer_vnet_name] = []
        zones_by_peer[external_zone.peer_vnet_name].append(external_zone)

    # Process zones by peer, aligning left and right at same Y levels
    for peer_vnet_name, peer_zones in zones_by_peer.items():
        right_zone = None
        left_zone = None
        for zone in peer_zones:
            if hasattr(zone, 'is_right_zone'):
                if zone.is_right_zone:
                    right_zone = zone
                else:
                    left_zone = zone
            else:
                peer_center_x = None
                if peer_vnet_name in hub_name_to_position:
                    peer_x, _, _ = hub_name_to_position[peer_vnet_name]
                    peer_center_x = peer_x + (VNET_WIDTH / 2)
                elif peer_vnet_name in regular_spoke_positions:
                    peer_x, _, _ = regular_spoke_positions[peer_vnet_name]
                    peer_center_x = peer_x + (VNET_WIDTH / 2)
                elif peer_vnet_name in hubless_spoke_positions:
                    peer_x, _, _ = hubless_spoke_positions[peer_vnet_name]
                    peer_center_x = peer_x + (VNET_WIDTH / 2)

                if peer_center_x:
                    zone_center_x = zone.zone_x + (VNET_WIDTH / 2)
                    if zone_center_x > peer_center_x:
                        right_zone = zone
                    else:
                        left_zone = zone
                else:
                    if right_zone is None:
                        right_zone = zone
                    else:
                        left_zone = zone

        # Get peer VNet position
        if peer_vnet_name in hub_name_to_position:
            peer_x, peer_y, peer_height = hub_name_to_position[peer_vnet_name]
        elif peer_vnet_name in regular_spoke_positions:
            peer_x, peer_y, peer_height = regular_spoke_positions[peer_vnet_name]
        elif peer_vnet_name in hubless_spoke_positions:
            peer_x, peer_y, peer_height = hubless_spoke_positions[peer_vnet_name]
        else:
            continue

        external_start_y = peer_y - EXTERNAL_SPACING_Y - VNET_MIN_HEIGHT
        max_count = max(
            len(right_zone.external_vnet_names) if right_zone else 0,
            len(left_zone.external_vnet_names) if left_zone else 0
        )

        for level in range(max_count):
            level_y = external_start_y - level * (VNET_MIN_HEIGHT + EXTERNAL_SPACING_Y)

            if right_zone and level < len(right_zone.external_vnet_names):
                ext_idx = level
                external_vnet_name = right_zone.external_vnet_names[ext_idx]
                peering_resource_id = right_zone.peering_resource_ids[ext_idx]

                group_id, external_main_id, cell_id_counter = create_external_vnet_group(
                    root, cell_id_counter, external_vnet_name, peering_resource_id, right_zone.zone_x, level_y
                )
                vnet_name_to_main_id[external_vnet_name] = external_main_id

                if peer_vnet_name in vnet_name_to_main_id:
                    peer_main_id = vnet_name_to_main_id[peer_vnet_name]
                    peer_vnet = next((v for v in topology.virtual_networks if v.name == peer_vnet_name), None)

                    if peer_vnet and external_vnet_name in peer_vnet.peering_names:
                        peering_index = peer_vnet.peering_names.index(external_vnet_name)
                        is_local = peer_vnet.peering_local_flags[peering_index] if peering_index < len(peer_vnet.peering_local_flags) else False
                        external_height_val = VNET_MIN_HEIGHT

                        if peer_vnet_name in regular_spoke_positions:
                            peer_x, peer_y, peer_height = regular_spoke_positions[peer_vnet_name]
                            _add_edge_if_not_exists(
                                peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                                float(peer_x), right_zone.zone_x, float(peer_y), level_y, float(peer_height), external_height_val
                            )
                        elif peer_vnet_name in hubless_spoke_positions:
                            peer_x, peer_y, peer_height = hubless_spoke_positions[peer_vnet_name]
                            _add_edge_if_not_exists(
                                peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                                float(peer_x), right_zone.zone_x, float(peer_y), level_y, float(peer_height), external_height_val
                            )
                        elif peer_vnet_name in hub_name_to_position:
                            peer_x, peer_y, peer_height = hub_name_to_position[peer_vnet_name]
                            _add_edge_if_not_exists(
                                peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                                float(peer_x), right_zone.zone_x, float(peer_y), level_y, float(peer_height), external_height_val
                            )

            if left_zone and level < len(left_zone.external_vnet_names):
                ext_idx = level
                external_vnet_name = left_zone.external_vnet_names[ext_idx]
                peering_resource_id = left_zone.peering_resource_ids[ext_idx]

                group_id, external_main_id, cell_id_counter = create_external_vnet_group(
                    root, cell_id_counter, external_vnet_name, peering_resource_id, left_zone.zone_x, level_y
                )
                vnet_name_to_main_id[external_vnet_name] = external_main_id

                if peer_vnet_name in vnet_name_to_main_id:
                    peer_main_id = vnet_name_to_main_id[peer_vnet_name]
                    peer_vnet = next((v for v in topology.virtual_networks if v.name == peer_vnet_name), None)

                    if peer_vnet and external_vnet_name in peer_vnet.peering_names:
                        peering_index = peer_vnet.peering_names.index(external_vnet_name)
                        is_local = peer_vnet.peering_local_flags[peering_index] if peering_index < len(peer_vnet.peering_local_flags) else False
                        external_height_val = VNET_MIN_HEIGHT

                        if peer_vnet_name in regular_spoke_positions:
                            peer_x, peer_y, peer_height = regular_spoke_positions[peer_vnet_name]
                            _add_edge_if_not_exists(
                                peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                                float(peer_x), left_zone.zone_x, float(peer_y), level_y, float(peer_height), external_height_val
                            )
                        elif peer_vnet_name in hubless_spoke_positions:
                            peer_x, peer_y, peer_height = hubless_spoke_positions[peer_vnet_name]
                            _add_edge_if_not_exists(
                                peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                                float(peer_x), left_zone.zone_x, float(peer_y), level_y, float(peer_height), external_height_val
                            )
                        elif peer_vnet_name in hub_name_to_position:
                            peer_x, peer_y, peer_height = hub_name_to_position[peer_vnet_name]
                            _add_edge_if_not_exists(
                                peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                                float(peer_x), left_zone.zone_x, float(peer_y), level_y, float(peer_height), external_height_val
                            )

    # Handle single external VNet zones
    for external_zone in external_zones:
        if external_zone.peer_vnet_name in zones_by_peer and len(zones_by_peer[external_zone.peer_vnet_name]) > 1:
            continue

        peer_vnet_name = external_zone.peer_vnet_name
        peer_x_pos2: float
        peer_y_pos2: float
        peer_height_val2: float
        peer_x_pos2, peer_y_pos2, peer_height_val2 = 0.0, 0.0, float(VNET_MIN_HEIGHT)
        if peer_vnet_name in hub_name_to_position:
            peer_x_pos2, peer_y_pos2, peer_height_val2 = hub_name_to_position[peer_vnet_name]
        elif peer_vnet_name in regular_spoke_positions:
            peer_x_pos2, peer_y_pos2, peer_height_val2 = regular_spoke_positions[peer_vnet_name]
        elif peer_vnet_name in hubless_spoke_positions:
            peer_x_pos2, peer_y_pos2, peer_height_val2 = hubless_spoke_positions[peer_vnet_name]
        else:
            continue

        external_start_y = peer_y - EXTERNAL_SPACING_Y - VNET_MIN_HEIGHT

        for idx, external_vnet_name in enumerate(external_zone.external_vnet_names):
            peering_resource_id = external_zone.peering_resource_ids[idx]
            y_position = external_start_y - idx * (VNET_MIN_HEIGHT + EXTERNAL_SPACING_Y)
            x_position = external_zone.zone_x

            group_id, external_main_id, cell_id_counter = create_external_vnet_group(
                root, cell_id_counter, external_vnet_name, peering_resource_id, x_position, y_position
            )
            vnet_name_to_main_id[external_vnet_name] = external_main_id

            if peer_vnet_name in vnet_name_to_main_id:
                peer_main_id = vnet_name_to_main_id[peer_vnet_name]
                peer_vnet = next((v for v in topology.virtual_networks if v.name == peer_vnet_name), None)

                if peer_vnet and external_vnet_name in peer_vnet.peering_names:
                    peering_index = peer_vnet.peering_names.index(external_vnet_name)
                    is_local = peer_vnet.peering_local_flags[peering_index] if peering_index < len(peer_vnet.peering_local_flags) else False
                    external_height_val = VNET_MIN_HEIGHT

                    if peer_vnet_name in regular_spoke_positions:
                        peer_x, peer_y, peer_height = regular_spoke_positions[peer_vnet_name]
                        _add_edge_if_not_exists(
                            peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                            float(peer_x), x_position, float(peer_y), y_position, float(peer_height), external_height_val
                        )
                    elif peer_vnet_name in hubless_spoke_positions:
                        peer_x, peer_y, peer_height = hubless_spoke_positions[peer_vnet_name]
                        _add_edge_if_not_exists(
                            peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                            float(peer_x), x_position, float(peer_y), y_position, float(peer_height), external_height_val
                        )
                    elif peer_vnet_name in hub_name_to_position:
                        peer_x, peer_y, peer_height = hub_name_to_position[peer_vnet_name]
                        _add_edge_if_not_exists(
                            peer_main_id, external_main_id, is_local, peer_vnet_name, external_vnet_name,
                            float(peer_x), x_position, float(peer_y), y_position, float(peer_height), external_height_val
                        )

    # PHASE 2: Create all edges AFTER all VNets
    for vnet in topology.virtual_networks:
        for i, peer_name in enumerate(vnet.peering_names):
            if peer_name in vnet_name_to_main_id and vnet.name in vnet_name_to_main_id:
                is_local = vnet.peering_local_flags[i] if i < len(vnet.peering_local_flags) else True
                source_main_id = vnet_name_to_main_id[vnet.name]
                target_main_id = vnet_name_to_main_id[peer_name]
                source_is_hub = vnet.name in hub_names and vnet.name not in hubless_spoke_names
                target_is_hub = peer_name in hub_names and peer_name not in hubless_spoke_names
                vnet_name_to_tenant_id.get(vnet.name)
                vnet_name_to_tenant_id.get(peer_name)

                source_x, source_y, source_height = None, None, None
                target_x, target_y, target_height = None, None, None

                if vnet.name in hub_name_to_position:
                    source_x, source_y, source_height = hub_name_to_position[vnet.name]
                elif vnet.name in regular_spoke_positions:
                    source_x, source_y, source_height = regular_spoke_positions[vnet.name]
                elif vnet.name in hubless_spoke_positions:
                    source_x, source_y, source_height = hubless_spoke_positions[vnet.name]
                elif vnet.name in all_vnet_positions:
                    source_x, source_y, source_height = all_vnet_positions[vnet.name]

                if peer_name in hub_name_to_position:
                    target_x, target_y, target_height = hub_name_to_position[peer_name]
                elif peer_name in regular_spoke_positions:
                    target_x, target_y, target_height = regular_spoke_positions[peer_name]
                elif peer_name in hubless_spoke_positions:
                    target_x, target_y, target_height = hubless_spoke_positions[peer_name]
                elif peer_name in all_vnet_positions:
                    target_x, target_y, target_height = all_vnet_positions[peer_name]

                if source_is_hub and target_is_hub and source_y is not None and source_height is not None and target_y is not None and target_height is not None:
                    source_center_y = source_y + (source_height / 2)
                    target_center_y = target_y + (target_height / 2)
                    _add_edge_if_not_exists(source_main_id, target_main_id, is_local, vnet.name, peer_name, source_x, target_x, source_center_y, target_center_y, source_height, target_height)
                elif source_x is not None and target_x is not None and source_y is not None and target_y is not None and source_height is not None and target_height is not None:
                    _add_edge_if_not_exists(source_main_id, target_main_id, is_local, vnet.name, peer_name, source_x, target_x, source_y, target_y, source_height, target_height)
                else:
                    _add_edge_if_not_exists(source_main_id, target_main_id, is_local, vnet.name, peer_name)

    return (
        cell_id_counter,
        vnet_name_to_main_id,
        hub_name_to_position,
        regular_spoke_positions,
        hubless_spoke_positions,
        all_vnet_positions,
        edges_to_create,
    )




"""Configuration loading and management for diagram generation."""

import logging
from pathlib import Path
from typing import Any

import yaml


class DiagramConfig:
    """Centralized configuration for diagram generation.

    Loads configuration from diagram_config.yaml and provides typed access
    to all configuration values with sensible defaults.
    """

    def __init__(self, config_path: Path | None = None):
        """Initialize configuration from YAML file.

        Args:
            config_path: Path to config file. If None, uses default location.
        """
        if config_path is None:
            config_path = Path(__file__).parent.parent / "diagram_config.yaml"

        try:
            with open(config_path, encoding="utf-8") as f:
                config = yaml.safe_load(f)
            self._layout = config.get("layout", {})
        except (FileNotFoundError, yaml.YAMLError) as e:
            logging.warning(f"Could not load diagram config from {config_path}: {e}. Using defaults.")
            self._layout = {}

    def _get(self, key: str, default: Any) -> Any:
        """Get config value with default."""
        return self._layout.get(key, default)

    def _get_nested(self, *keys: str, default: Any) -> Any:
        """Get nested config value with default.

        Args:
            *keys: Path to nested value (e.g., "icon", "vnet_size")
            default: Default value if not found

        Returns:
            Config value or default
        """
        value = self._layout
        for key in keys:
            if isinstance(value, dict):
                value = value.get(key)
                if value is None:
                    return default
            else:
                return default
        return value if value is not None else default

    # Canvas settings
    @property
    def canvas_padding(self) -> int:
        return int(self._get("canvas_padding", 20))

    @property
    def page_width(self) -> int:
        return int(self._get("page_width", 1169))

    @property
    def page_height(self) -> int:
        return int(self._get("page_height", 827))

    # Zone settings
    @property
    def zone_width(self) -> int:
        return int(self._get("zone_width", 920))

    @property
    def zone_spacing(self) -> int:
        return int(self._get("zone_spacing", 200))

    # VNet box dimensions
    @property
    def vnet_width(self) -> int:
        return int(self._get("vnet_width", 400))

    @property
    def vnet_height_base(self) -> int:
        return int(self._get("vnet_height_base", 50))

    @property
    def vnet_min_height(self) -> int:
        return int(self._get("vnet_min_height", 70))

    # Spacing constants
    @property
    def vnet_spacing_x(self) -> int:
        return int(self._get("vnet_spacing_x", 300))

    @property
    def vnet_spacing_y(self) -> int:
        return int(self._get("vnet_spacing_y", 100))

    @property
    def hub_y(self) -> int:
        return int(self._get("hub_y", 200))

    @property
    def spoke_start_y(self) -> int:
        return int(self._get("spoke_start_y", 300))

    @property
    def spacing_below_hub(self) -> int:
        return int(self._get("spacing_below_hub", 50))

    @property
    def spacing_between_spokes(self) -> int:
        return int(self._get("spacing_between_spokes", 20))

    # Subnet layout
    @property
    def subnet_padding_top(self) -> int:
        return int(self._get("subnet_padding_top", 75))

    @property
    def subnet_height(self) -> int:
        return int(self._get("subnet_height", 30))

    @property
    def subnet_spacing(self) -> int:
        return int(self._get("subnet_spacing", 5))

    @property
    def subnet_padding_left(self) -> int:
        return int(self._get_nested("subnet", "padding_left", default=5))

    @property
    def subnet_width_offset(self) -> int:
        return int(self._get_nested("subnet", "width_offset", default=10))

    @property
    def subnet_padding_bottom(self) -> int:
        return int(self._get_nested("subnet", "padding_bottom", default=10))

    # External VNet positioning
    @property
    def external_spacing_x(self) -> int:
        return int(self._get("external_spacing_x", 50))

    @property
    def external_spacing_y(self) -> int:
        return int(self._get("external_spacing_y", 100))

    @property
    def stub_spacing_x(self) -> int:
        return int(self._get("stub_spacing_x", 100))

    # Icon settings
    @property
    def icon_vnet_size(self) -> int:
        return int(self._get_nested("icon", "vnet_size", default=20))

    @property
    def icon_subnet_size(self) -> int:
        return int(self._get_nested("icon", "subnet_size", default=20))

    @property
    def icon_subnet_height(self) -> int:
        return int(self._get_nested("icon", "subnet_height", default=12))

    @property
    def icon_spacing(self) -> int:
        return int(self._get_nested("icon", "icon_spacing", default=5))

    @property
    def icon_margin_right(self) -> int:
        return int(self._get_nested("icon", "icon_margin_right", default=6))

    # Icon image paths
    @property
    def icon_vnet(self) -> str:
        return str(self._get_nested("icon", "vnet", default="img/lib/azure2/networking/Virtual_Networks.svg"))

    @property
    def icon_subnet(self) -> str:
        return str(self._get_nested("icon", "subnet", default="img/lib/azure2/networking/Subnet.svg"))

    @property
    def icon_route_table(self) -> str:
        return str(self._get_nested("icon", "route_table", default="img/lib/azure2/networking/Route_Tables.svg"))

    @property
    def icon_nsg(self) -> str:
        return str(self._get_nested("icon", "nsg", default="img/lib/azure2/networking/Network_Security_Groups.svg"))

    @property
    def icon_firewall(self) -> str:
        return str(self._get_nested("icon", "firewall", default="img/lib/azure2/networking/Firewalls.svg"))

    @property
    def icon_bastion(self) -> str:
        return str(self._get_nested("icon", "bastion", default="img/lib/azure2/networking/Bastions.svg"))

    @property
    def icon_vpn_gateway(self) -> str:
        return str(self._get_nested("icon", "vpn_gateway", default="img/lib/mscae/VPN_Gateway.svg"))

    @property
    def icon_expressroute(self) -> str:
        return str(self._get_nested("icon", "expressroute", default="img/lib/azure2/networking/ExpressRoute_Circuits.svg"))

    @property
    def icon_router_vpn(self) -> str:
        return str(self._get_nested("icon", "router_vpn", default="img/lib/allied_telesis/security/Router_VPN.svg"))

    @property
    def icon_router_expressroute(self) -> str:
        return str(self._get_nested("icon", "router_expressroute", default="img/lib/active_directory/shadowed_router.svg"))

    @property
    def icon_nat(self) -> str:
        return str(self._get_nested("icon", "nat", default="img/lib/azure2/networking/NAT.svg"))

    @property
    def icon_private_endpoint(self) -> str:
        return str(self._get_nested("icon", "private_endpoint", default="img/lib/azure2/other/Private_Endpoints.svg"))

    @property
    def icon_ddos(self) -> str:
        return str(self._get_nested("icon", "ddos", default="img/lib/azure2/networking/DDoS_Protection_Plans.svg"))

    # Routing/waypoint settings
    @property
    def routing_waypoint_offset_horizontal(self) -> int:
        return int(self._get_nested("routing", "waypoint_offset_horizontal", default=50))

    @property
    def routing_waypoint_offset_vertical(self) -> int:
        return int(self._get_nested("routing", "waypoint_offset_vertical", default=20))

    @property
    def routing_waypoint_offset_close(self) -> int:
        return int(self._get_nested("routing", "waypoint_offset_close", default=5))

    @property
    def routing_clearance(self) -> int:
        return int(self._get_nested("routing", "route_clearance", default=20))

    # Style colors
    @property
    def color_hub_stroke(self) -> str:
        return str(self._get_nested("colors", "hub_stroke", default="#0078D4"))

    @property
    def color_hub_fill(self) -> str:
        return str(self._get_nested("colors", "hub_fill", default="#E6F1FB"))

    @property
    def color_hub_font(self) -> str:
        return str(self._get_nested("colors", "hub_font", default="#004578"))

    @property
    def color_spoke_stroke(self) -> str:
        return str(self._get_nested("colors", "spoke_stroke", default="#CC6600"))

    @property
    def color_spoke_fill(self) -> str:
        return str(self._get_nested("colors", "spoke_fill", default="#f2f7fc"))

    @property
    def color_spoke_font(self) -> str:
        return str(self._get_nested("colors", "spoke_font", default="#CC6600"))

    @property
    def color_external_stroke(self) -> str:
        return str(self._get_nested("colors", "external_stroke", default="#6C757D"))

    @property
    def color_external_fill(self) -> str:
        return str(self._get_nested("colors", "external_fill", default="#F8F9FA"))

    @property
    def color_external_font(self) -> str:
        return str(self._get_nested("colors", "external_font", default="#495057"))

    @property
    def color_stub_stroke(self) -> str:
        return str(self._get_nested("colors", "stub_stroke", default="#20B2AA"))

    @property
    def color_stub_fill(self) -> str:
        return str(self._get_nested("colors", "stub_fill", default="#E0F7F5"))

    @property
    def color_stub_font(self) -> str:
        return str(self._get_nested("colors", "stub_font", default="#006666"))

    @property
    def color_subnet_stroke(self) -> str:
        return str(self._get_nested("colors", "subnet_stroke", default="#C8C6C4"))

    @property
    def color_subnet_fill(self) -> str:
        return str(self._get_nested("colors", "subnet_fill", default="#FAF9F8"))

    @property
    def color_subnet_font(self) -> str:
        return str(self._get_nested("colors", "subnet_font", default="#323130"))

    @property
    def color_hub_to_hub(self) -> str:
        return str(self._get_nested("colors", "hub_to_hub", default="#6A1B9A"))

    @property
    def color_hub_to_spoke(self) -> str:
        return str(self._get_nested("colors", "hub_to_spoke", default="#0078D4"))

    # Edge colors (from existing config)
    @property
    def edge_color_hubless_spoke(self) -> str:
        return str(self._get("edge_color_hubless_spoke", "#8B0000"))

    @property
    def edge_color_cross_tenant(self) -> str:
        return str(self._get("edge_color_cross_tenant", "#808080"))

    # Edge line patterns
    @property
    def edge_pattern_hub_to_hub(self) -> str:
        return str(self._get_nested("edge_patterns", "hub_to_hub", default="solid"))

    @property
    def edge_pattern_hub_to_spoke(self) -> str:
        return str(self._get_nested("edge_patterns", "hub_to_spoke", default="solid"))

    @property
    def edge_pattern_hubless_spoke(self) -> str:
        return str(self._get_nested("edge_patterns", "hubless_spoke", default="dashed"))

    @property
    def edge_pattern_cross_tenant(self) -> str:
        return str(self._get_nested("edge_patterns", "cross_tenant", default="solid"))

    @property
    def edge_pattern_external(self) -> str:
        return str(self._get_nested("edge_patterns", "external", default="solid"))

    @property
    def edge_dash_pattern(self) -> str:
        return str(self._get_nested("edge_patterns", "dash_pattern", default="8 8"))

    # Legend settings
    @property
    def legend_width(self) -> int:
        return int(self._get("legend_width", 250))

    @property
    def legend_height(self) -> int:
        return int(self._get("legend_height", 600))

    @property
    def legend_padding(self) -> int:
        return int(self._get("legend_padding", 15))

    @property
    def legend_item_spacing(self) -> int:
        return int(self._get("legend_item_spacing", 8))

    @property
    def legend_item_height(self) -> int:
        return int(self._get("legend_item_height", 20))

    @property
    def legend_x(self) -> int:
        return int(self._get("legend_x", 20))

    @property
    def legend_y(self) -> int:
        return int(self._get("legend_y", 50))

    @property
    def legend_to_diagram_spacing(self) -> int:
        return int(self._get("legend_to_diagram_spacing", 20))

    # Legend layout settings (3-column layout)
    @property
    def legend_column_gap(self) -> int:
        return int(self._get_nested("legend", "column_gap", default=15))

    @property
    def legend_column_margin(self) -> int:
        return int(self._get_nested("legend", "column_margin", default=5))

    @property
    def legend_vnet_box_width(self) -> int:
        return int(self._get_nested("legend", "vnet_box_width", default=40))

    @property
    def legend_vnet_box_height(self) -> int:
        return int(self._get_nested("legend", "vnet_box_height", default=30))

    @property
    def legend_vnet_box_to_label_gap(self) -> int:
        return int(self._get_nested("legend", "vnet_box_to_label_gap", default=45))

    @property
    def legend_connection_line_length(self) -> int:
        return int(self._get_nested("legend", "connection_line_length", default=50))

    @property
    def legend_connection_line_to_label_gap(self) -> int:
        return int(self._get_nested("legend", "connection_line_to_label_gap", default=5))

    @property
    def legend_icon_size(self) -> int:
        return int(self._get_nested("legend", "icon_size", default=20))

    @property
    def legend_icons_per_sub_column(self) -> int:
        return int(self._get_nested("legend", "icons_per_sub_column", default=4))

    @property
    def legend_icon_sub_column_gap(self) -> int:
        return int(self._get_nested("legend", "icon_sub_column_gap", default=5))

    @property
    def legend_icon_to_label_gap(self) -> int:
        return int(self._get_nested("legend", "icon_to_label_gap", default=5))

    # Cell ID management
    @property
    def cell_id_start(self) -> int:
        return int(self._get_nested("cell_id", "start", default=10))

    # Draw.io settings
    @property
    def drawio_grid_size(self) -> int:
        return int(self._get_nested("drawio", "grid_size", default=10))

    @property
    def drawio_page_dx(self) -> int:
        return int(self._get_nested("drawio", "page_dx", default=1422))

    @property
    def drawio_page_dy(self) -> int:
        return int(self._get_nested("drawio", "page_dy", default=794))

    # Label/text spacing
    @property
    def label_spacing_top(self) -> int:
        return int(self._get_nested("label", "spacing_top", default=5))

    @property
    def label_spacing_left(self) -> int:
        return int(self._get_nested("label", "spacing_left", default=5))

    # Hybrid connectivity settings
    @property
    def hybrid_onprem_x_offset(self) -> int:
        return int(self._get_nested("hybrid", "onprem_x_offset", default=200))

    @property
    def hybrid_azure_x(self) -> int:
        return int(self._get_nested("hybrid", "azure_x", default=700))

    @property
    def hybrid_connection_spacing(self) -> int:
        return int(self._get_nested("hybrid", "connection_spacing", default=300))

    @property
    def hybrid_vertical_spacing(self) -> int:
        return int(self._get_nested("hybrid", "vertical_spacing", default=150))

    @property
    def hybrid_onprem_box_width(self) -> int:
        return int(self._get_nested("hybrid", "onprem_box_width", default=250))

    @property
    def hybrid_onprem_box_height(self) -> int:
        return int(self._get_nested("hybrid", "onprem_box_height", default=140))

    @property
    def hybrid_router_icon_size(self) -> int:
        return int(self._get_nested("hybrid", "router_icon_size", default=30))

    @property
    def hybrid_content_start_y(self) -> int:
        return int(self._get_nested("hybrid", "content_start_y", default=35))

    @property
    def hybrid_content_spacing(self) -> int:
        return int(self._get_nested("hybrid", "content_spacing", default=8))

    @property
    def hybrid_content_bottom_padding(self) -> int:
        return int(self._get_nested("hybrid", "content_bottom_padding", default=10))

    @property
    def hybrid_info_box_width(self) -> int:
        return int(self._get_nested("hybrid", "info_box_width", default=300))

    @property
    def hybrid_info_box_line_height(self) -> int:
        return int(self._get_nested("hybrid", "info_box_line_height", default=20))

    @property
    def hybrid_info_box_padding(self) -> int:
        return int(self._get_nested("hybrid", "info_box_padding", default=10))

    @property
    def hybrid_info_box_offset_y(self) -> int:
        return int(self._get_nested("hybrid", "info_box_offset_y", default=50))


# Global config instance (singleton pattern)
_config_instance: DiagramConfig | None = None


def get_config() -> DiagramConfig:
    """Get the global configuration instance."""
    global _config_instance
    if _config_instance is None:
        _config_instance = DiagramConfig()
    return _config_instance

